"""Utilities for the local runtime models."""

import zipfile

import pyaxmlparser
import plistlib


def get_package_name(path: str) -> str:
    """Get the package name of an android file.

    Args:
        path: Path to the android file.

    Returns:
        str: Package name of the android file.
    """

    apk_info = pyaxmlparser.APK(path)
    return apk_info.get_package()


def get_bundle_id(path: str) -> str:
    """Get the bundle identifier of an iOS ipa file.

    Args:
        path: Path to the iOS ipa file.

    Returns:
        str: Bundle identifier of the iOS ipa file.
    """

    with zipfile.ZipFile(path, "r") as z:
        # Locate the Info.plist file within the IPA
        infoplist = None
        for filename in z.namelist():
            dirs = filename.split("/")
            # payload, Payload and payLoad the three formats can be found
            if (
                len(dirs) == 3
                and dirs[0].lower().strip() == "payload"
                and dirs[2].lower().strip() == "info.plist"
            ):
                content = z.open(filename).read()
                infoplist = plistlib.loads(content)

        if infoplist is None:
            return ""

        # Extract and read the Info.plist file
        bundle_id = infoplist.get("CFBundleIdentifier")
        if bundle_id is None:
            return ""

        return bundle_id



"""Agent asset."""

import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.agent")
class Agent(asset.Asset):
    """Agent asset."""

    key: str
    version: Optional[str] = None
    git_location: Optional[str] = None
    docker_location: Optional[str] = None
    yaml_file_location: Optional[str] = None

    def __str__(self) -> str:
        if self.version is not None:
            return f"Agent {self.key}:{self.version}"
        else:
            return f"Agent {self.key}"

    @property
    def proto_field(self) -> str:
        return "agent"



"""Lists full details of a single vulnerability or all the vulnerabilities for a specific scan."""

import json
from typing import Dict, Optional

from ostorlab.apis import request


class ScanVulnzDescribeAPIRequest(request.APIRequest):
    """Lists vulnerabilities of a scan."""

    def __init__(
        self,
        scan_id: int,
        vuln_id: int = None,
        page: int = 1,
        number_elements: int = 10,
    ):
        self._scan_id = scan_id
        self._vuln_id = vuln_id
        self._page = page
        self._page = page
        self._number_elements = number_elements

    @property
    def query(self) -> Optional[str]:
        """Defines the query to list the vulnz.

        Returns:
            The query to list the vulnz.
        """
        return """
            query Vulnerabilities(
                $scanId: Int!
                $vulnerabilityId: Int
                $page: Int!
                $numberElements: Int!
              ) {
                scan(scanId: $scanId) {
                  isEditable
                  vulnerabilities(
                    vulnerabilityId: $vulnerabilityId
                    page: $page
                    numberElements: $numberElements
                  ) {
                    pageInfo {
                      hasNext
                      hasPrevious
                      count
                      numPages
                    }
                    vulnerabilities {
                      id
                      technicalDetail
                      technicalDetailFormat
                      customRiskRating
                      customCvssV3BaseScore
                      falsePositive
                      vulnerabilityLocation
                      detail {
                        title
                        shortDescription
                        description
                        recommendation
                        cvssV3Vector
                        references {
                          title
                          url
                        }
                      }
                    }
                  }
                }
            }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query to list the vulnz.

        Returns:
              The query to list the vulnz.
        """
        data = {
            "query": self.query,
            "variables": json.dumps(
                {
                    "scanId": self._scan_id,
                    "vulnerabilityId": self._vuln_id,
                    "page": self._page,
                    "numberElements": self._number_elements,
                }
            ),
        }
        return data



"""Utils to generate and manipulate strings."""

import json
import random
import string
from typing import Any


def random_string(length: int, alphabet: str = string.ascii_lowercase) -> str:
    """Generates a random string of a specified length.

    Args:
        length: Length of the string to generate.
        alphabet: The alphabet used to generate the random string, defaults to ascii lower case.

    Returns:
        Randomly generated string.
    """
    if length <= 0:
        raise ValueError(f"Invalid string length: {length}")

    result = "".join((random.choice(alphabet) for _ in range(length)))
    return result


def _format_data(data: Any) -> Any:
    if isinstance(data, dict):
        return {k: _format_data(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_format_data(v) for v in data]
    elif isinstance(data, bytes):
        return f"<bytes of length {len(data)}>"
    elif isinstance(data, str):
        if len(data) > 4096:
            return f"{data[:256]}... <string of length {len(data)}>"
        else:
            return data
    else:
        return data


def format_dict(dict_obj: dict[str, Any]) -> str:
    """Format message for logging, filtering out large values."""
    filtered_data = _format_data(dict_obj)
    return json.dumps(filtered_data, indent=2)



"""Asset of type .AAB package file.
This module takes care of preparing a file of type .AAB before calling the create mobile scan API.
"""

import io
import click

from ostorlab.cli.ci_scan.run import run
from ostorlab.cli.ci_scan.run.assets import mobile
from ostorlab.apis import scan_create as scan_create_api


@run.run.command()
@click.argument("file", type=click.File(mode="rb"), required=True)
@click.pass_context
def android_aab(ctx: click.core.Context, file: io.FileIO) -> None:
    """Create scan for android .APK package file."""
    mobile.run_mobile_scan(ctx, file, scan_create_api.MobileAssetType.ANDROID)



"""Logger is in charge of defining the output format based on CircleCI expected output."""

import os

from ostorlab.cli import console as cli_console
from ostorlab.cli.ci_scan.run.ci_logger import logger

console = cli_console.Console()


class Logger(logger.Logger):
    """Logger is in charge of printing the results based on the CI expected output."""

    def info(self, message: str) -> None:
        """Print Info messages.

        Args:
            message: message to print.
        """
        console.info(message)

    def error(self, message: str) -> None:
        """Print Error messages.

        Args:
            message: message to print.
        """
        console.error(message)

    def output(self, name: str, value: str) -> None:
        """Pass output to the next step of the CI by setting them as an env variable.

        Args:
            name: name of the output to pass to the next step.
            value: value of the output.
        """
        os.environ[name.upper()] = value



"""IPv4 address asset."""

import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.ip.v4")
class IPv4(asset.Asset):
    """IPv4 Address target asset."""

    host: str
    version: int = 4
    mask: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.host}/{self.mask}"

    @property
    def proto_field(self) -> str:
        return "ipv4"



"""This module is the entry point for OXO CLI."""

import logging
import json
from typing import Optional

import click

from ostorlab import configuration_manager

logger = logging.getLogger("CLI")


@click.group()
@click.pass_context
@click.version_option()
@click.option("--api-key", help="API key to login to the platform.", required=False)
@click.option("--proxy", "-X", help="Proxy to route HTTPS requests through.")
@click.option(
    "--tlsverify/--no-tlsverify",
    help="Control TLS server certificate verification.",
    default=True,
)
@click.option("-d", "--debug/--no-debug", help="Enable debug mode", default=False)
@click.option("-v", "--verbose/--no-verbose", help="Enable verbose mode", default=False)
@click.option(
    "--gcp-logging-credential",
    type=click.Path(exists=True),
    help="Path to GCP logging JSON credential file.",
    required=False,
)
def rootcli(
    ctx: click.core.Context,
    proxy: Optional[str] = None,
    tlsverify: Optional[bool] = True,
    debug: bool = False,
    verbose: bool = False,
    api_key: str = None,
    gcp_logging_credential: Optional[str] = None,
) -> None:
    """Oxo is an open-source project to help automate security testing.\n
    Oxo standardizes interoperability between tools in a consistent, scalable, and performant way.
    """
    ctx.obj = {}
    ctx.obj["proxy"] = proxy
    ctx.obj["tlsverify"] = tlsverify
    ctx.obj["api_key"] = api_key
    # Configuration is a singleton class. One initiated with the provided API key, others will make use of it.
    conf_manager = configuration_manager.ConfigurationManager()
    conf_manager.api_key = api_key
    ctx.obj["config_manager"] = conf_manager

    if verbose is True:
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for current_logger in loggers:
            current_logger.setLevel(logging.INFO)
    if debug is True:
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for current_logger in loggers:
            current_logger.setLevel(logging.DEBUG)
    if gcp_logging_credential is not None:
        try:
            import google.cloud.logging
            from google.oauth2 import service_account

            with open(gcp_logging_credential, "r", encoding="utf-8") as source:
                content = source.read()
                info = json.loads(content)

            credentials = service_account.Credentials.from_service_account_info(info)
            client = google.cloud.logging.Client(credentials=credentials)
            client.setup_logging()
            ctx.obj["gcp_logging_credential"] = content
        except ImportError:
            logger.error(
                "Could not import Google Cloud Logging, install it with `pip install 'ostorlab[google-cloud-logging]'"
            )



"""Ostorlab lite local runtime modules. this module takes care of launching a scan in the
local machine but with little setup, like no MQ service, no scan persistence and no defaults
agents are started."""

from .runtime import LiteLocalRuntime

__all__ = [LiteLocalRuntime]



"""Module for the command list inside the group scan.
This module takes care of listing all the remote or local scans.
Example of usage:
    - ostorlab scan list --source=source."""

from typing import Tuple

import click
from ostorlab.cli.scan import scan
from ostorlab.cli import console as cli_console

console = cli_console.Console()


@scan.command()
@click.argument("scan_ids", nargs=-1, type=int, required=False)
@click.option(
    "--all",
    "-a",
    "stop_all",
    is_flag=True,
    help="Stop all running scans",
    default=False,
)
@click.pass_context
def stop(ctx: click.core.Context, scan_ids: Tuple[int, ...], stop_all: bool) -> None:
    """Stop one or multiple scans.\n
    Usage:\n
        - ostorlab scan --runtime=local stop 4
        - ostorlab scan --runtime=local stop 4 5 6
        - ostorlab scan --runtime=local stop --all
    """
    if len(scan_ids) == 0 and stop_all is False:
        raise click.UsageError("Either provide scan IDs or use --all flag")

    runtime_instance = ctx.obj["runtime"]
    if stop_all is True:
        scans_list = runtime_instance.list()
        ids_to_stop = [
            s.id
            for s in scans_list
            if s.progress == "in_progress" or s.progress == "not_started"
        ]
        if len(ids_to_stop) == 0:
            console.warning("No running scans found.")
            return
    else:
        ids_to_stop = list(scan_ids)

    console.info(f"Stopping {len(ids_to_stop)} scan(s).")
    for scan_id in ids_to_stop:
        runtime_instance.stop(scan_id=scan_id)



"""Domain name asset definition."""

import dataclasses

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.domain_name")
class DomainName(asset.Asset):
    """Domain Name target asset per RFC 1034 and 1035."""

    name: str

    def __str__(self) -> str:
        return self.name

    @property
    def proto_field(self) -> str:
        return "domain_name"



"""Display the agent install progress."""

from typing import Dict, Iterator

from rich import progress

from ostorlab.cli import console as cli_console


console = cli_console.Console()


class AgentInstallProgress(progress.Progress):
    """Class responsible for displaying the progress of the agent installation."""

    def __init__(self):
        super().__init__(
            "[progress.description]{task.description}",
            progress.BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            progress.DownloadColumn(),
            "•",
            progress.TransferSpeedColumn(),
            "•",
            progress.TimeRemainingColumn(),
        )

    def display(self, logs_generator: Iterator[Dict]) -> None:
        """Display the progress of the agent install command.

        Args:
            logs_generator: generator of the agent install command logs.
        """
        download_tasks = {}
        extract_tasks = {}

        with self as prg:
            for log in logs_generator:
                if "Downloading" in log.get("status", []):
                    task_id = log["id"]
                    if task_id not in download_tasks:
                        task = prg.add_task(
                            f"[red]Download : {task_id}",
                            total=log.get("progressDetail", {}).get("total"),
                        )
                        download_tasks[task_id] = task
                    task = download_tasks[task_id]
                    current_progress = log.get("progressDetail", {}).get("current")
                    prg.update(task, completed=current_progress)

                elif "Download complete" in log.get("status", []):
                    task_id = log["id"]
                    if task_id in download_tasks:
                        task = download_tasks[task_id]
                        prg.update(task, total=100.0)

                elif "Extracting" in log.get("status", []):
                    task_id = log["id"]
                    if task_id not in extract_tasks:
                        task = prg.add_task(
                            f"[red]Extract : {task_id}",
                            total=log.get("progressDetail", {}).get("total"),
                        )
                        extract_tasks[task_id] = task
                    task = extract_tasks[task_id]
                    current_progress = log.get("progressDetail", {}).get("current")
                    prg.update(task, completed=current_progress)

        console.success("Installation successful")



"""MQ Mixin.

Definition of the main methods to publish and consume MQ messages by the agents.
"""

import asyncio
import concurrent.futures
import logging
from typing import List, Optional

import aio_pika
import tenacity
from aiormq import exceptions as aiormq_exceptions

logger = logging.getLogger(__name__)
NUMBER_RETRIES = 6
WAIT_FIXED_TIME = 5


class AgentMQMixin:
    """MQ Mixin class used to initialize the channel, send messages and process them."""

    def __init__(
        self,
        name: str,
        keys: List[str],
        url: str,
        topic: str,
        max_priority: Optional[int] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """Initialize the MQ parameters, the channel pools and the executors to process the messages.
        Args:
            name: Name of the queue.
            keys: Selectors that the queue listens to.
            url: string for the MQ url might contain ssl parameters e.g.`amqps://user:pass@host//`
            topic: string with exchange name
            max_priority: Optional To declare a priority queue, it is a positive int [1, 255],
             indicating the max priority the queue supports.
        """
        self._name = name
        self._keys = keys
        self._queue_name = f"{self._name}_queue"
        self._url = url
        self._topic = topic
        self._loop = loop or asyncio.get_event_loop()
        self._max_priority = max_priority
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)
        self._queue: Optional[aio_pika.Queue] = None
        self._connection_pool: aio_pika.pool.Pool[aio_pika.Connection] = (
            aio_pika.pool.Pool(self._get_connection, max_size=32, loop=self._loop)
        )
        self._channel_pool: aio_pika.pool.Pool[aio_pika.Channel] = aio_pika.pool.Pool(
            self._get_channel, max_size=64, loop=self._loop
        )

    async def _get_connection(self) -> aio_pika.abc.AbstractRobustConnection:
        return await aio_pika.connect_robust(
            url=self._url, loop=self._loop, fail_fast=False
        )

    async def _get_channel(self) -> aio_pika.Channel:
        async with self._connection_pool.acquire() as connection:
            channel: aio_pika.Channel = await connection.channel()
            return channel

    async def _get_exchange(
        self, channel: aio_pika.abc.AbstractChannel
    ) -> aio_pika.abc.AbstractExchange:
        return await channel.declare_exchange(
            self._topic,
            type=aio_pika.ExchangeType.TOPIC,
            arguments={"x-max-length": 10000, "x-overflow": "reject-publish"},
        )

    async def mq_init(self, delete_queue_first: bool = False) -> None:
        """Declares the queue to start preparing for receiving messages.
        Args:
            delete_queue_first: Used for testing purposes. To delete pending queues first.
        """
        logger.info("Connecting to %s", self._url)
        connection = await self._get_connection()
        async with connection:
            channel = await connection.channel()
            await self._declare_mq_queue(channel, delete_queue_first)

    async def mq_run(self, delete_queue_first: bool = False) -> None:
        """Use a channel to declare the queue, set the listener on the selectors and consume the received messaged.
        Args:
            delete_queue_first: Used for testing purposes. To delete pending queues first.
        """
        connection = await self._get_connection()
        channel = await connection.channel()
        await self._declare_mq_queue(channel, delete_queue_first)

        if self._queue is None:
            raise ValueError("queue not created")

        await self._queue.consume(self._mq_process_message, no_ack=False)

    async def _declare_mq_queue(
        self,
        channel: aio_pika.abc.AbstractRobustChannel,
        delete_queue_first: bool = False,
    ) -> None:
        """Declare the MQ queue on a given channel.
        The queue is durable, redeclaring the queue will return the same queue
        Args:
            channel: the MQ channel to use for the queues
            delete_queue_first: Used for testing purposes. To delete pending queues first.
        """
        await channel.set_qos(prefetch_count=1)
        exchange = await self._get_exchange(channel)
        if delete_queue_first:
            await channel.queue_delete(self._queue_name)

        if self._max_priority is not None:
            self._queue = await channel.declare_queue(
                self._queue_name,
                auto_delete=False,
                durable=True,
                arguments={"x-max-priority": self._max_priority},
            )
        else:
            self._queue = await channel.declare_queue(
                self._queue_name, auto_delete=False, durable=True
            )

        if self._queue is None:
            raise ValueError("queue not declared.")

        for k in self._keys:
            await self._queue.bind(exchange, k)

    async def _mq_process_message(
        self, message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:
        """Consumes the MQ messages and calls the process message callback."""
        logger.debug("incoming pika message received")
        try:
            async with message.process(requeue=True, reject_on_redelivered=True):
                await self._loop.run_in_executor(
                    self._executor,
                    self.process_message,
                    message.routing_key,
                    message.body,
                )
        except aio_pika.exceptions.ChannelInvalidStateError:
            logger.warning("The channel is closed unexpectedly.")
            await self.mq_run()

    def process_message(self, selector: str, message: bytes) -> None:
        """Callback to implement to process the MQ messages received."""
        raise NotImplementedError()

    async def async_mq_send_message(
        self, key: str, message: bytes, message_priority: Optional[int] = None
    ) -> None:
        """Async Send the message to the provided routing key and its priority.
        Args:
            message: Message to send
            key: Selector where to send the message.
            message_priority: the priority of the message. Default is 0
        """
        logger.debug("sending %s to %s", message, key)
        async with self._channel_pool.acquire() as channel:
            exchange = await self._get_exchange(channel)
            pika_message = aio_pika.Message(
                body=message,
                priority=message_priority or 0,
                delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
            )
            await exchange.publish(routing_key=key, message=pika_message)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            (
                aio_pika.exceptions.ConnectionClosed,
                ConnectionResetError,
                concurrent.futures.CancelledError,
                aiormq_exceptions.ChannelInvalidStateError,
                aiormq_exceptions.AMQPConnectionError,
            )
        ),
        stop=tenacity.stop_after_attempt(NUMBER_RETRIES),
        wait=tenacity.wait_fixed(WAIT_FIXED_TIME),
        reraise=True,
    )
    def mq_send_message(
        self, key: str, message: bytes, message_priority: Optional[int] = None
    ) -> None:
        """The method sends the message to the selected key with the defined priority in async mode .
        Args:
            key: Selector that the queue listens to.
            message: Message to send.
            message_priority: the priority to use for the message default is 0.
        """
        logger.debug("sending %s to %s", message, key)

        if not self._loop.is_running():
            self._loop.run_until_complete(
                self.async_mq_send_message(key, message, message_priority)
            )
        else:
            future = asyncio.run_coroutine_threadsafe(
                self.async_mq_send_message(key, message, message_priority), self._loop
            )
            try:
                result = future.result()
            except Exception as exc:
                logger.warning("The coroutine raised an exception: %s", exc)
                raise
            else:
                logger.debug("The coroutine returned: %s", result)



"""Handles calls to the token API.

Typical usage example:
public_runner = LoginAPIRunner(username, password, otp_token)
public_runner.login_user()
"""

import httpx
from typing import Dict, Optional, Any

from ostorlab.apis.runners import runner
from ostorlab.apis import request as api_request
from ostorlab.apis import login


TOKEN_ENDPOINT = "https://api.ostorlab.co/apis/token/"


class LoginAPIRunner(runner.APIRunner):
    """Responsible for the public API calls, and preparing the responses."""

    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        otp_token: Optional[str] = None,
        proxy: Optional[str] = None,
        verify: bool = True,
    ):
        """Constructs all the necessary attributes for the object.

        Args:
            username: the username (email) used to log in.
            password: the password used to log in.
            proxy: The proxy through which a request is made. Defaults to None.
            verify: Whether to verify the TLS certificate. Defaults to True.
        """
        super().__init__(proxy, verify)
        self._username = username
        self._password = password
        self._otp_token = otp_token

    @property
    def endpoint(self) -> str:
        """Token API endpoint."""
        return TOKEN_ENDPOINT

    def login_user(self) -> httpx.Response:
        """Logs in the user.

        Returns:
            The API response.
        """
        if self._username is None or self._password is None:
            raise ValueError("Missing credentials.")

        login_request = login.UsernamePasswordLoginAPIRequest(
            self._username, self._password, self._otp_token
        )

        return self._sent_request(login_request)

    def execute(self, request: api_request.APIRequest) -> Dict[str, Any]:
        """Executes a request using the Token GraphQL API.

        Args:
            request: The request to be executed

        Raises:
            ResponseError: When the API returns an error

        Returns:
            The API response
        """
        response = self._sent_request(request)
        if response.status_code != 200:
            raise runner.ResponseError(
                f"Response status code is {response.status_code}: {response.content.decode(errors='ignore')}"
            )
        data: Dict[str, Any] = response.json()
        errors = data.get("errors")
        if errors is not None and isinstance(errors, list):
            error = errors[0].get("message")
            raise runner.ResponseError(f"Response errors: {error}")
        else:
            return data



"""Create Agent Group via an API Request."""

from typing import Dict, List, Optional, Union

from ostorlab.apis import request

AgentType = Dict[str, Union[str, List]]


class CreateAgentGroupAPIRequest(request.APIRequest):
    """Persist agent group API request."""

    def __init__(
        self, name: Optional[str], description: str, agents: List[AgentType]
    ) -> None:
        """Initializer"""
        self._name = name
        self._description = description
        self._agents = agents

    @property
    def query(self) -> Optional[str]:
        """Sets the query of the API request.

        Returns:
            The query to create the agent group.
        """
        return """
            mutation PublishAgentGroup($agentGroup: AgentGroupCreateInputType!){
                publishAgentGroup(agentGroup: $agentGroup) {
                    agentGroup{
                        id
                    }
                }
            }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request.

        Returns:
            The body of the create agent group request.
        """

        variables = {
            "agentGroup": {
                "name": self._name,
                "description": self._description,
                "access": "PRIVATE",
                "agents": self._agents,
            }
        }
        data = {"query": self.query, "variables": variables}
        return data



"""Common logic for Mobile Assets of type .APK package file AAB package file and IPA .
This module takes care of preparing the application file and calling the create mobile scan API.
"""

import io
import itertools
import json
from typing import List, Optional

import click

from ostorlab.cli.ci_scan.run import run
from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner as base_runner
from ostorlab.apis import scan_create as scan_create_api
from ostorlab.apis import test_credentials_create as test_credentials_create_api


def _prepare_test_credentials(
    ctx: click.core.Context,
) -> List[test_credentials_create_api.TestCredential]:
    test_credentials_login = ctx.obj["test_credentials"]["test_credentials_login"]
    test_credentials_password = ctx.obj["test_credentials"]["test_credentials_password"]
    test_credentials_url = ctx.obj["test_credentials"]["test_credentials_url"]
    test_credentials_role = ctx.obj["test_credentials"]["test_credentials_role"]
    test_credentials_name = ctx.obj["test_credentials"]["test_credentials_name"]
    test_credentials_value = ctx.obj["test_credentials"]["test_credentials_value"]

    credentials = []
    for login, password, role, url in itertools.zip_longest(
        test_credentials_login,
        test_credentials_password,
        test_credentials_role,
        test_credentials_url,
    ):
        credentials.append(
            test_credentials_create_api.TestCredentialLogin(
                login=login, password=password, role=role, url=url
            )
        )

    if test_credentials_name:
        credentials.append(
            test_credentials_create_api.TestCredentialCustom(
                values=dict(
                    itertools.zip_longest(test_credentials_name, test_credentials_value)
                )
            )
        )

    return credentials


def run_mobile_scan(
    ctx: click.core.Context,
    file: io.FileIO,
    asset_type: scan_create_api.MobileAssetType,
) -> None:
    """Create scan for mobile application package file."""
    ci_logger = ctx.obj["ci_logger"]
    if ctx.obj.get("api_key"):
        scan_profile = ctx.obj["scan_profile"]
        title = ctx.obj["title"]
        break_on_risk_rating = ctx.obj["break_on_risk_rating"]
        max_wait_minutes = ctx.obj["max_wait_minutes"]
        sboms = ctx.obj["sboms"]
        repository = ctx.obj["repository"]
        source = ctx.obj["source"]
        pr_number = ctx.obj["pr_number"]
        branch = ctx.obj["branch"]
        scope_urls_regexes = ctx.obj["scope_urls_regexes"]
        ui_prompt_ids = ctx.obj.get("ui_prompt_ids") or []
        ui_prompt_names = ctx.obj.get("ui_prompt_names") or []
        ui_prompt_actions = ctx.obj.get("ui_prompt_actions") or []
        runner = authenticated_runner.AuthenticatedAPIRunner(
            api_key=ctx.obj.get("api_key")
        )
        try:
            test_credentials = _prepare_test_credentials(ctx)
            if test_credentials:
                credential_ids = _create_test_credentials(test_credentials, runner)
                ci_logger.info(
                    f"Created test credentials {', '.join(str(t) for t in test_credentials)} successfully"
                )
            else:
                credential_ids = []

            ui_automation_rule_ids: List[int] = []

            if len(ui_prompt_ids) > 0:
                ui_automation_rule_ids.extend(ui_prompt_ids)
                ci_logger.info(f"Using existing UI prompts with IDs: {ui_prompt_ids}")

            if len(ui_prompt_names) > 0 and len(ui_prompt_actions) > 0:
                ui_prompts_json = [
                    {"name": name, "code": action}
                    for name, action in zip(ui_prompt_names, ui_prompt_actions)
                ]
                try:
                    ci_logger.info("Creating UI prompts...")
                    prompts_result = runner.execute(
                        scan_create_api.CreateUIPromptsAPIRequest(
                            ui_prompts=ui_prompts_json
                        )
                    )
                    created_prompt_ids = [
                        int(prompt["id"])
                        for prompt in prompts_result["data"]["createUiPrompts"][
                            "uiPrompts"
                        ]
                    ]
                    ui_automation_rule_ids.extend(created_prompt_ids)
                    ci_logger.info(f"Created UI prompts with IDs: {created_prompt_ids}")
                except (json.JSONDecodeError, KeyError) as e:
                    ci_logger.error(
                        f"Invalid UI prompts format: {e}. Continuing without UI prompts."
                    )

            ci_logger.info(
                f"creating scan `{title}` with profile `{scan_profile}` for `{asset_type}`"
            )
            scan_source = None
            if source is not None:
                scan_source = scan_create_api.ScanSource(
                    source=source,
                    repository=repository,
                    pr_number=pr_number,
                    branch=branch,
                )
            if scope_urls_regexes is not None and len(scope_urls_regexes) > 0:
                scope_urls_regexes = list(scope_urls_regexes)
            else:
                scope_urls_regexes = None

            scan_id = _create_scan(
                title=title,
                scan_profile=scan_profile,
                asset_type=asset_type,
                file=file,
                credential_ids=credential_ids,
                runner=runner,
                sboms=sboms,
                scan_source=scan_source,
                scope_urls_regexes=scope_urls_regexes,
                ui_automation_rule_ids=ui_automation_rule_ids,
            )

            ci_logger.output(name="scan_id", value=scan_id)
            ci_logger.info(f"Scan created with id {scan_id}.")

            if break_on_risk_rating is not None and break_on_risk_rating != "":
                run.apply_break_scan_risk_rating(
                    break_on_risk_rating, scan_id, max_wait_minutes, runner, ci_logger
                )

        except base_runner.ResponseError as e:
            ci_logger.error(f"Could not start the scan. {e}")
            raise click.exceptions.Exit(2) from e
        except TimeoutError:
            ci_logger.error("The scan is still running.")
            raise click.exceptions.Exit(2) from None
    else:
        ci_logger.error("API key not not provided.")
        raise click.exceptions.Exit(2) from None


def _create_scan(
    title: str,
    scan_profile: str,
    asset_type: scan_create_api.MobileAssetType,
    file: io.FileIO,
    credential_ids: List[int],
    runner: authenticated_runner.AuthenticatedAPIRunner,
    sboms: List[io.FileIO],
    scan_source: Optional[scan_create_api.ScanSource] = None,
    scope_urls_regexes: Optional[List[str]] = None,
    ui_automation_rule_ids: List[int] = (),
) -> int:
    scan_result = runner.execute(
        scan_create_api.CreateMobileScanAPIRequest(
            title=title,
            asset_type=asset_type,
            scan_profile=scan_profile,
            application=file,
            test_credential_ids=credential_ids,
            sboms=sboms,
            scan_source=scan_source,
            scope_urls_regexes=scope_urls_regexes,
            ui_automation_rule_ids=ui_automation_rule_ids,
        )
    )
    scan_id = scan_result.get("data").get("createMobileScan").get("scan").get("id")
    return scan_id


def _create_test_credentials(test_credentials, runner):
    credential_ids = []
    for test_credential in test_credentials:
        test_credential_result = runner.execute(
            test_credentials_create_api.CreateTestCredentialAPIRequest(
                test_credential=test_credential
            )
        )
        credential_ids.append(
            test_credential_result.get("data")
            .get("createTestCredentials")
            .get("testCredentials")
            .get("id")
        )
    return credential_ids



"""Create mobile scan API."""

import enum
import io
import json
import dataclasses
from typing import Dict, Optional, BinaryIO, List, Any

from . import request


class CreateUIPromptsAPIRequest(request.APIRequest):
    """Create UI prompts API request with get_or_create logic."""

    def __init__(self, ui_prompts: List[Dict[str, Any]]):
        self._ui_prompts = ui_prompts

    @property
    def query(self) -> str:
        """Defines the query to create UI prompts.

        Returns:
            The query to create UI prompts
        """
        return """
mutation CreateUIPrompts($uiPrompts: [UIAutomationRulesInputType]!) {
  createUiPrompts(uiPrompts: $uiPrompts) {
    uiPrompts {
      id
      name
      code
    }
  }
}
        """

    @property
    def data(self) -> Dict[str, Any]:
        """Sets the query and variables to create UI prompts.

        Returns:
            The query and variables to create UI prompts.
        """
        return {
            "query": self.query,
            "variables": json.dumps({"uiPrompts": self._ui_prompts}),
        }

    @property
    def is_json(self) -> bool:
        """Indicates that the request should be sent as JSON.

        Returns:
            True if the request should be sent as JSON.
        """
        return True


@dataclasses.dataclass
class ScanSource:
    """Dataclass holding scan source related parameters."""

    source: str
    repository: Optional[str] = None
    pr_number: Optional[str] = None
    branch: Optional[str] = None


SCAN_PROFILES = {
    "fast_scan": "Fast Scan",
    "full_scan": "Full Scan",
    "full_web_scan": "Full Web Scan",
    # aliases
    "fast": "Fast Scan",
    "full": "Full Scan",
    "full_web": "Full Web Scan",
}


class MobileAssetType(enum.Enum):
    ANDROID = enum.auto()
    IOS = enum.auto()


class CreateMobileScanAPIRequest(request.APIRequest):
    """Create mobile scan API from a file."""

    def __init__(
        self,
        title: str,
        asset_type: MobileAssetType,
        scan_profile: str,
        application: BinaryIO,
        test_credential_ids: Optional[List[int]] = None,
        scope_urls_regexes: Optional[List[str]] = None,
        sboms: list[io.FileIO] = None,
        scan_source: Optional[ScanSource] = None,
        ui_automation_rule_ids: List[int] = (),
    ):
        self._title = title
        self._asset_type = asset_type
        self._scan_profile = scan_profile
        self._application = application
        self._test_credential_ids = test_credential_ids
        self._scope_urls_regexes = scope_urls_regexes
        self._sboms = sboms
        self._scan_source = scan_source
        self._ui_automation_rule_ids = ui_automation_rule_ids

    @property
    def query(self) -> Optional[str]:
        """Defines the query to create a mobile scan.

        Returns:
            The query to create a mobile scan
        """

        return """
mutation MobileScan($title: String!, $assetType: String!, $application: Upload!, $sboms: [Upload!], $scanProfile: String!, $credentialIds: [Int], $scanSource: ScanSourceInputType, $scopeUrlsRegexes: [String], $uiAutomationRuleInstances: [UIAutomationRuleInstanceInputType]) {
  createMobileScan(title: $title, assetType: $assetType, application: $application, sboms: $sboms, scanProfile: $scanProfile, credentialIds: $credentialIds, scanSource: $scanSource, scopeUrlsRegexes: $scopeUrlsRegexes, uiAutomationRuleInstances: $uiAutomationRuleInstances) {
    scan {
      id
    }
  }
}
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query and variables to create the scan.

        Returns:
            The query and variables to create a scan.
        """

        var_map = {"0": ["variables.application"]}
        for idx, _ in enumerate(self._sboms):
            var_map[str(idx + 1)] = [f"variables.sboms.{idx}"]

        data = {
            "operations": json.dumps(
                {
                    "query": self.query,
                    "variables": {
                        "title": self._title,
                        "assetType": self._asset_type.name.lower(),
                        "application": None,
                        "scanProfile": self._scan_profile,
                        "credentialIds": self._test_credential_ids,
                        "sboms": [None for _ in self._sboms],
                        "scanSource": {
                            "source": self._scan_source.source,
                            "repository": self._scan_source.repository,
                            "prNumber": self._scan_source.pr_number,
                            "branch": self._scan_source.branch,
                        }
                        if self._scan_source is not None
                        else None,
                        "scopeUrlsRegexes": self._scope_urls_regexes,
                        "uiAutomationRuleInstances": [
                            {
                                "ruleId": rule_id,
                                "args": [{"name": f"Rule {rule_id}"}],
                            }
                            for rule_id in self._ui_automation_rule_ids
                        ],
                    },
                }
            ),
            "map": json.dumps(var_map),
        }
        return data

    @property
    def files(self) -> Optional[Dict]:
        """Sets the file for multipart upload to create the mobile scan.

        Returns:
            The file mapping to create a scan.
        """
        files = {"0": self._application}
        for idx, sbom_file in enumerate(self._sboms):
            files[str(idx + 1)] = sbom_file
        return files


class CreateWebScanAPIRequest(request.APIRequest):
    """Create web scan API from a file."""

    def __init__(
        self,
        title: str,
        urls: List[str],
        scan_profile: str,
        sboms: Optional[list[io.FileIO]] = None,
        api_schema: Optional[io.FileIO] = None,
        proxy: Optional[str] = None,
        qps: Optional[int] = None,
        filtered_url_regexes: Optional[List[str]] = None,
        test_credential_ids: Optional[List[int]] = None,
        ui_automation_rule_ids: List[int] = (),
    ):
        self._title = title
        self._urls = urls
        self._scan_profile = scan_profile
        self._api_schema = api_schema
        self._sboms = sboms
        self._proxy = proxy
        self._qps = qps
        self._filtered_url_regexes = filtered_url_regexes
        self._test_credential_ids = test_credential_ids
        self._ui_automation_rule_ids = ui_automation_rule_ids

    @property
    def query(self) -> Optional[str]:
        """Defines the query to create a web scan.

        Returns:
            The query to create a web scan
        """

        return """
mutation WebScan($title: String!, $urls: [String]!, $scanProfile: String!, $sboms: [Upload!], $apiSchema: Upload, $proxy: String, $qps: Int, $filteredUrlRegexes: [String], $credentialIds: [Int], $uiAutomationRuleInstances: [UIAutomationRuleInstanceInputType]) {
  createWebScan(title: $title, urls: $urls, scanProfile: $scanProfile, sboms: $sboms, apiSchema: $apiSchema, proxy: $proxy, qps: $qps, filteredUrlRegexes: $filteredUrlRegexes, credentialIds: $credentialIds, uiAutomationRuleInstances: $uiAutomationRuleInstances) {
    scan {
      id
    }
  }
}
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query and variables to create the scan.

        Returns:
            The query and variables to create a scan.
        """

        var_map = {}
        for idx, _ in enumerate(self._sboms):
            var_map[str(idx)] = [f"variables.sboms.{idx}"]

        if self._api_schema is not None:
            var_map[str(len(self._sboms))] = ["variables.apiSchema"]

        if len(var_map) > 0:
            data = {
                "operations": json.dumps(
                    {
                        "query": self.query,
                        "variables": {
                            "title": self._title,
                            "urls": self._urls,
                            "scanProfile": self._scan_profile,
                            "apiSchema": None,
                            "credentialIds": self._test_credential_ids,
                            "filteredUrlRegexes": self._filtered_url_regexes,
                            "proxy": self._proxy,
                            "qps": self._qps,
                            "sboms": [None for _ in self._sboms],
                            "uiAutomationRuleInstances": [
                                {
                                    "ruleId": rule_id,
                                    "args": [{"name": f"Rule {rule_id}"}],
                                }
                                for rule_id in self._ui_automation_rule_ids
                            ],
                        },
                    }
                ),
                "map": json.dumps(var_map),
            }
        else:
            data = {
                "query": self.query,
                "variables": json.dumps(
                    {
                        "title": self._title,
                        "urls": self._urls,
                        "scanProfile": self._scan_profile,
                        "apiSchema": None,
                        "credentialIds": self._test_credential_ids,
                        "filteredUrlRegexes": self._filtered_url_regexes,
                        "proxy": self._proxy,
                        "qps": self._qps,
                        "sboms": [None for _ in self._sboms],
                        "uiAutomationRuleInstances": [
                            {
                                "ruleId": rule_id,
                                "args": [{"name": f"Rule {rule_id}"}],
                            }
                            for rule_id in self._ui_automation_rule_ids
                        ],
                    }
                ),
            }
        return data

    @property
    def files(self) -> Optional[Dict]:
        """Sets the file for multipart upload to create the web scan.

        Returns:
            The file mapping to create a scan.
        """
        files = {}
        for idx, sbom_file in enumerate(self._sboms):
            files[str(idx)] = sbom_file

        if self._api_schema is not None:
            files[str(len(self._sboms))] = self._api_schema

        return files



"""Add exploitation and post exploitation details

Revision ID: c3231845f850
Revises: d47f712dab53
Create Date: 2025-03-26 10:59:07.006256

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "c3231845f850"
down_revision = "d47f712dab53"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("agent_argument", schema=None) as batch_op:
        batch_op.alter_column(
            "value",
            existing_type=sa.TEXT(),
            type_=sa.LargeBinary(),
            existing_nullable=True,
        )

    with op.batch_alter_table("vulnerability", schema=None) as batch_op:
        batch_op.add_column(sa.Column("exploitation_detail", sa.Text(), nullable=True))
        batch_op.add_column(
            sa.Column("post_exploitation_detail", sa.Text(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("vulnerability", schema=None) as batch_op:
        batch_op.drop_column("post_exploitation_detail")
        batch_op.drop_column("exploitation_detail")

    with op.batch_alter_table("agent_argument", schema=None) as batch_op:
        batch_op.alter_column(
            "value",
            existing_type=sa.LargeBinary(),
            type_=sa.TEXT(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



"""Jaeger service to collect Agent traces."""

import logging
from typing import Dict, Optional

import docker
import tenacity
from docker import errors
from docker import types
from docker.models import services

logger = logging.getLogger(__name__)

JAEGER_IMAGE = "jaegertracing/all-in-one:latest"
DEFAULT_EXPOSED_PORTS = {
    5775: 5775,
    6831: 6831,
    6832: 6832,
    5778: 5778,
    16686: 16686,
    14268: 14268,
    9411: 9411,
}
DEFAULT_JAEGER_PORT = 6831
DEFAULT_JAEGER_UI_PORT = 16686


class LocalJaeger:
    """Jaeger service spawned a docker swarm service."""

    def __init__(
        self,
        name: str,
        network: str,
        exposed_ports: Dict[int, int] = None,
        image: str = JAEGER_IMAGE,
    ) -> None:
        """Initialize the Jaeger service parameters.
        Args:
            name: Name of the service.
            network: Network used for the Docker Jaeger service.
            exposed_ports: The list of Jaeger service exposed ports
            image: Jaeger Docker image
        """
        self._name = name
        self._docker_client = docker.from_env()
        # images
        self._jaeger_image = image
        self._network = network
        self._jaeger_host = f"jaeger_{self._name}"
        # service
        self._jaeger_service = None
        # exposed_port
        self._exposed_ports = exposed_ports or DEFAULT_EXPOSED_PORTS

    @property
    def url(self) -> str:
        """URL to connect to the local Jaeger instance."""
        return f"jaeger://{self._jaeger_host}:{DEFAULT_JAEGER_PORT}"

    @property
    def management_interface_ui(self) -> str:
        """URL to management user interface of the Jaeger instance."""
        return f"http://127.0.0.1:{DEFAULT_JAEGER_UI_PORT}"

    @property
    def service(self):
        return self._jaeger_service

    def start(self) -> None:
        """Start local Jaeger instance."""
        self._create_network()
        self._jaeger_service = self._start_jaeger()

    def stop(self):
        """Stop the local Jaeger instance."""
        for service in self._docker_client.services.list():
            universe = service.attrs["Spec"]["Labels"].get("ostorlab.universe")
            if (
                universe is not None
                and service.name.startswith("jaeger_")
                and self._name in universe
            ):
                service.remove()

    def _create_network(self):
        if any(
            network.name == self._network
            for network in self._docker_client.networks.list()
        ):
            logger.warning("network already exists.")
        else:
            logger.info("creating private network %s", self._network)
            return self._docker_client.networks.create(
                name=self._network,
                driver="overlay",
                attachable=True,
                labels={"ostorlab.universe": self._name},
                check_duplicate=True,
            )

    def _start_jaeger(self) -> Optional[services.Service]:
        try:
            logger.info("starting Jaeger")
            endpoint_spec = types.services.EndpointSpec(
                mode="vip", ports=self._exposed_ports
            )
            service_mode = types.services.ServiceMode("replicated", replicas=1)
            return self._docker_client.services.create(
                image=self._jaeger_image,
                networks=[self._network],
                name=self._jaeger_host,
                restart_policy=types.RestartPolicy(condition="any"),
                mode=service_mode,
                labels={"ostorlab.universe": self._name, "ostorlab.jaeger": ""},
                endpoint_spec=endpoint_spec,
            )
        except docker.errors.APIError as e:
            error_message = f"Jaeger service could not be started. Reason: {e}."
            logger.error(error_message)
            return

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(20),
        wait=tenacity.wait_fixed(0.5),
        # return last value and don't raise RetryError exception.
        retry_error_callback=lambda lv: lv.outcome,
        retry=tenacity.retry_if_result(lambda v: v is False),
    )
    def is_service_healthy(self) -> bool:
        logger.info("checking service %s", self._jaeger_service.name)
        return self.is_healthy

    @property
    def is_healthy(self) -> bool:
        try:
            return (
                self._jaeger_service is not None
                and len(
                    [
                        task
                        for task in self._jaeger_service.tasks()
                        if task["Status"]["State"] == "running"
                    ]
                )
                == 1
            )
        except errors.DockerException:
            return False



"""API Schema asset."""

import dataclasses
from typing import Optional

from ostorlab.assets import asset

PROTO_FIELD = "api_schema"
CLASS_NAME = "API Schema"


@dataclasses.dataclass
@asset.selector("v3.asset.file.api_schema")
class ApiSchema(asset.Asset):
    """API Schema target asset."""

    endpoint_url: str
    content: Optional[bytes] = None
    content_url: Optional[str] = None
    schema_type: Optional[str] = None

    def __str__(self) -> str:
        str_representation = CLASS_NAME
        if self.schema_type is not None:
            str_representation = f"{str_representation} ({self.schema_type})"
        str_representation = f"{str_representation}: {self.endpoint_url}"

        return str_representation

    @property
    def proto_field(self) -> str:
        return PROTO_FIELD



"""Module responsible for fetching the agent details from the container image."""

import logging
import io
from typing import Any, Optional

import docker
import docker.errors

from ostorlab import configuration_manager
from ostorlab.apis import agent_details as agent_details_api
from ostorlab.apis.runners import public_runner, authenticated_runner
from ostorlab.apis.runners import runner as base_runner
from ostorlab.cli import console as cli_console
from ostorlab.agent import definitions as agent_definitions
from ostorlab.utils import version as version_definition

console = cli_console.Console()
logger = logging.getLogger(__name__)


class Error(Exception):
    """Base Error."""


class AgentDetailsNotFound(Error):
    """Agent not found error."""


def get_details(agent_key: str) -> dict[str, Any]:
    """Sends an API request with the agent key, and retrieve the agent information.

    Args:
        agent_key: the agent key in the form : agent/org/name

    Returns:
        dictionary of the agent information like : name, dockerLocation..

    Raises:
        AgentDetailsNotFound: If the agent is not found.
    """
    config_manager = configuration_manager.ConfigurationManager()

    if config_manager.is_authenticated is True:
        runner = authenticated_runner.AuthenticatedAPIRunner()
    else:
        runner = public_runner.PublicAPIRunner()

    try:
        response = runner.execute(agent_details_api.AgentDetailsAPIRequest(agent_key))
    except base_runner.ResponseError as e:
        raise AgentDetailsNotFound("requested agent not found") from e

    if "errors" in response:
        error_message = f"""The provided agent key : {agent_key} does not correspond to any agent.
        Please make sure you have the correct agent key.
        """
        raise AgentDetailsNotFound(error_message)
    else:
        agent_details = response["data"]["agent"]
        return agent_details


def get_definition(
    agent_key: str,
    version: Optional[str] = None,
) -> agent_definitions.AgentDefinition:
    """Fetch the agent definition.

    Args:
         agent_key: key of the agent in agent/org/agentName format.
         version: Optional version of the agent. If None, uses latest available.

    Returns:
        agent_definition: AgentDefinition object containing the agent definition.

    Raises:
         AgentDetailsNotFound: If the agent image is not found.
    """
    image_name = get_container_image(agent_key=agent_key, version=version)
    if image_name is None:
        raise AgentDetailsNotFound(f"Agent {agent_key} not found")
    client = docker.from_env()
    docker_image = client.images.get(image_name)
    yaml_definition_string = docker_image.labels.get("agent_definition")
    with io.StringIO(yaml_definition_string) as file:
        agent_definition = agent_definitions.AgentDefinition.from_yaml(file)

    return agent_definition


def get_container_image(agent_key: str, version: Optional[str] = None) -> Optional[str]:
    """Get the agent container image name based on the agent key and optional version. If the version is not specified,
    the latest version is returned if found, otherwise, None is returned.

    Args:
        agent_key : The agent key.
        version : The version of the container image.

    Returns:
        The matching container image name with version. None if no matching image found.
    """
    image = agent_key.replace("/", "_")
    logger.debug("Searching container name %s with version %s", image, version)
    client = docker.from_env()
    matching_tag_versions = []
    for img in client.images.list(name=image):
        for t in img.tags:
            splitted_tag = t.split(":")
            if len(splitted_tag) == 2:
                t_name, t_tag = splitted_tag
            else:
                t_name = ":".join(splitted_tag[:-1])
                t_tag = splitted_tag[-1]
            if t_name == image and version is None:
                try:
                    matching_tag_versions.append(version_definition.Version(t_tag[1:]))
                except ValueError:
                    logger.warning("Invalid version %s", t_tag[1:])
            elif t_name == image and version is not None:
                if t_tag[1:] == version:
                    try:
                        matching_tag_versions.append(
                            version_definition.Version(t_tag[1:])
                        )
                    except ValueError:
                        logger.warning("Invalid version %s", t_tag[1:])

    if len(matching_tag_versions) == 0:
        return None

    matching_version = max(matching_tag_versions)
    return f"{image}:v{matching_version}"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/report/risk/risk.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ostorlab.agent.message.proto.v3.asset.domain_name import domain_name_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_domain__name_dot_domain__name__pb2
from ostorlab.agent.message.proto.v3.asset.ip.v4 import v4_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_ip_dot_v4_dot_v4__pb2
from ostorlab.agent.message.proto.v3.asset.ip.v6 import v6_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_ip_dot_v6_dot_v6__pb2
from ostorlab.agent.message.proto.v3.asset.link import link_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_link_dot_link__pb2
from ostorlab.agent.message.proto.v3.asset.store.android_store import android_store_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_store_dot_android__store_dot_android__store__pb2
from ostorlab.agent.message.proto.v3.asset.store.ios_store import ios_store_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_store_dot_ios__store_dot_ios__store__pb2
from ostorlab.agent.message.proto.v3.asset.file import file_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_file_dot_file__pb2
from ostorlab.agent.message.proto.v3.asset.file.android.aab import aab_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_file_dot_android_dot_aab_dot_aab__pb2
from ostorlab.agent.message.proto.v3.asset.file.android.apk import apk_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_file_dot_android_dot_apk_dot_apk__pb2
from ostorlab.agent.message.proto.v3.asset.file.ios.ipa import ipa_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_file_dot_ios_dot_ipa_dot_ipa__pb2
from ostorlab.agent.message.proto.v3.asset.file.api_schema import api_schema_pb2 as ostorlab_dot_agent_dot_message_dot_proto_dot_v3_dot_asset_dot_file_dot_api__schema_dot_api__schema__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n6ostorlab/agent/message/proto/v3/report/risk/risk.proto\x12+ostorlab.agent.message.proto.v3.report.risk\x1a\x43ostorlab/agent/message/proto/v3/asset/domain_name/domain_name.proto\x1a\x34ostorlab/agent/message/proto/v3/asset/ip/v4/v4.proto\x1a\x34ostorlab/agent/message/proto/v3/asset/ip/v6/v6.proto\x1a\x35ostorlab/agent/message/proto/v3/asset/link/link.proto\x1aMostorlab/agent/message/proto/v3/asset/store/android_store/android_store.proto\x1a\x45ostorlab/agent/message/proto/v3/asset/store/ios_store/ios_store.proto\x1a\x35ostorlab/agent/message/proto/v3/asset/file/file.proto\x1a@ostorlab/agent/message/proto/v3/asset/file/android/aab/aab.proto\x1a@ostorlab/agent/message/proto/v3/asset/file/android/apk/apk.proto\x1a<ostorlab/agent/message/proto/v3/asset/file/ios/ipa/ipa.proto\x1a\x46ostorlab/agent/message/proto/v3/asset/file/api_schema/api_schema.proto\"\xa8\x07\n\x07Message\x12Q\n\x0b\x64omain_name\x18\x01 \x01(\x0b\x32:.ostorlab.agent.message.proto.v3.asset.domain_name.MessageH\x00\x12\x44\n\x04ipv4\x18\x02 \x01(\x0b\x32\x34.ostorlab.agent.message.proto.v3.asset.ip.v4.MessageH\x00\x12\x44\n\x04ipv6\x18\x03 \x01(\x0b\x32\x34.ostorlab.agent.message.proto.v3.asset.ip.v6.MessageH\x00\x12\x43\n\x04link\x18\x04 \x01(\x0b\x32\x33.ostorlab.agent.message.proto.v3.asset.link.MessageH\x00\x12[\n\randroid_store\x18\x05 \x01(\x0b\x32\x42.ostorlab.agent.message.proto.v3.asset.store.android_store.MessageH\x00\x12S\n\tios_store\x18\x06 \x01(\x0b\x32>.ostorlab.agent.message.proto.v3.asset.store.ios_store.MessageH\x00\x12\x43\n\x04\x66ile\x18\x07 \x01(\x0b\x32\x33.ostorlab.agent.message.proto.v3.asset.file.MessageH\x00\x12V\n\x0b\x61ndroid_aab\x18\x08 \x01(\x0b\x32?.ostorlab.agent.message.proto.v3.asset.file.android.aab.MessageH\x00\x12V\n\x0b\x61ndroid_apk\x18\t \x01(\x0b\x32?.ostorlab.agent.message.proto.v3.asset.file.android.apk.MessageH\x00\x12N\n\x07ios_ipa\x18\n \x01(\x0b\x32;.ostorlab.agent.message.proto.v3.asset.file.ios.ipa.MessageH\x00\x12T\n\napi_schema\x18\r \x01(\x0b\x32>.ostorlab.agent.message.proto.v3.asset.file.api_schema.MessageH\x00\x12\x13\n\x0b\x64\x65scription\x18\x0b \x02(\t\x12\x0e\n\x06rating\x18\x0c \x02(\tB\x07\n\x05\x61sset')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.report.risk.risk_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=807
  _MESSAGE._serialized_end=1743
# @@protoc_insertion_point(module_scope)



"""Asset of type file.
This module takes care of preparing a generic file of type before injecting it to the runtime instance."""

import io
import logging
from typing import Optional, Tuple

import click

from ostorlab.assets import file as file_asset
from ostorlab.cli import console as cli_console
from ostorlab.cli.scan.run import run
from ostorlab import exceptions

console = cli_console.Console()
logger = logging.getLogger(__name__)


@run.run.command(name="file")
@click.option("--file", type=click.File(mode="rb"), multiple=True, required=False)
@click.option("--url", multiple=True, required=False)
@click.pass_context
def file_cli(
    ctx: click.core.Context,
    file: Optional[Tuple[io.FileIO]] = None,
    url: Optional[Tuple[str]] = None,
) -> None:
    """Run scan for file asset."""
    runtime = ctx.obj["runtime"]
    assets = []
    if len(file) > 0:
        for f in file:
            assets.append(file_asset.File(content=f.read(), path=str(f.name)))
    elif len(url) > 0:
        for u in url:
            assets.append(file_asset.File(content_url=u))
    else:
        console.error("Command accepts either path or source url of the file.")
        raise click.exceptions.Exit(2)

    logger.debug("scanning assets %s", [str(asset) for asset in assets])
    try:
        runtime.scan(
            title=ctx.obj["title"],
            agent_group_definition=ctx.obj["agent_group_definition"],
            assets=assets,
        )
    except exceptions.OstorlabError as e:
        console.error(f"An error was encountered while running the scan: {e}")



"""Vulnz module that handles listing and describing a vulnerability."""

import click

from ostorlab.cli.rootcli import rootcli
from ostorlab.runtimes import registry


@rootcli.group()
@click.option(
    "--runtime",
    type=click.Choice(["local", "litelocal", "cloud", "hybrid"]),
    help="""Runtime is in charge of preparing the environment to trigger a scan.\n
                    Specify which runtime to use: \n
                    local: on you local machine\n
                    litelocal: stripped down local runtime\n
                    cloud: on Ostorlab cloud, (requires login)\n
                    hybrid: running partially on Ostorlab cloud and partially on the local machine\n
                   """,
    default="local",
    required=True,
)
@click.pass_context
def vulnz(ctx, runtime: str = "local") -> None:
    """You can use vulnz to list and describe vulnerabilities.\n"""
    try:
        runtime_instance = registry.select_runtime(
            runtime_type=runtime,
            gcp_logging_credential=ctx.obj.get("gcp_logging_credential"),
        )
        ctx.obj["runtime"] = runtime_instance
    except registry.RuntimeNotFoundError as e:
        raise click.ClickException(
            f"The selected runtime {runtime} is not supported."
        ) from e
    pass



"""Create test credentials API."""

import abc
import dataclasses
import json
from typing import Dict, Optional, Any

from . import request


class TestCredential(abc.ABC):
    """Base abstract test credentials."""

    @abc.abstractmethod
    def to_variables(self) -> Dict[str, Any]:
        """Generate query variables."""
        pass


@dataclasses.dataclass
class TestCredentialLogin(TestCredential):
    """Login password test credentials with optional role and url fields."""

    login: str
    password: str
    role: Optional[str] = None
    url: Optional[str] = None

    def to_variables(self) -> Dict[str, Any]:
        """Generate query variables."""
        return {
            "testCredentials": {
                "loginPassword": {
                    "login": self.login,
                    "password": self.password,
                    "role": self.role,
                    "url": self.url,
                }
            }
        }

    def __repr__(self) -> str:
        return f"login: {self.login}, password: ************, role: {self.role}, url: {self.url}"


@dataclasses.dataclass
class TestCredentialCustom(TestCredential):
    """Custom test credentials with variable number of a pair of name and value."""

    values: Dict[str, str]

    def to_variables(self) -> Dict[str, Any]:
        """Generate query variables."""
        return {
            "testCredentials": {
                "custom": {
                    "credentials": [
                        {"name": n, "value": v} for (n, v) in self.values.items()
                    ]
                }
            }
        }


class CreateTestCredentialAPIRequest(request.APIRequest):
    """Create mobile scan API from a file."""

    def __init__(self, test_credential: TestCredential):
        self._test_credential = test_credential

    @property
    def query(self) -> Optional[str]:
        """Defines the query to create test credentials.

        Returns:
            The query to create a test credential.
        """

        return """
mutation TestCredentials($testCredentials: TestCredentialsInput!) {
  createTestCredentials(testCredentials: $testCredentials) {
    testCredentials {
      ... on CustomTestCredentials {
        id
      }
      ... on LoginPasswordTestCredentials {
        id
      }
    }
  }
}
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query and variables to create test credentials.

        Returns:
            The query and variables to create test credentials.
        """
        data = {
            "query": self.query,
            "variables": json.dumps(self._test_credential.to_variables()),
        }
        return data



"""IP address asset."""

import dataclasses
import ipaddress
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.ip")
class IP(asset.Asset):
    """IP Address target asset."""

    host: str
    version: Optional[int] = None
    mask: Optional[str] = None

    def __post_init__(self) -> None:
        if self.version is None:
            self.version = ipaddress.ip_address(self.host).version

    def __str__(self) -> str:
        return f"{self.host}/{self.mask}"



"""Module for the command list inside the group scan.
This module takes care of listing all the remote or local scans.
Example of usage:
    - ostorlab scan list --source=source."""

from typing import Union

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.scan import scan
from ostorlab.runtimes.local.models import models
from ostorlab.utils import styles, risk_rating

console = cli_console.Console()


@scan.command(name="list")
@click.option(
    "--page", "-p", help="Page number of scans you would like to see.", default=1
)
@click.option("--elements", "-e", help="Number of scans to show per page.", default=10)
@click.pass_context
def list_scans(ctx: click.core.Context, page: int, elements: int) -> None:
    """List all your scans.\n
    Usage:\n
        - ostorlab scan --runtime=runtime list
    """
    runtime_instance = ctx.obj["runtime"]
    with console.status("Fetching scans"):
        scans = runtime_instance.list(page=page, number_elements=elements)
        if scans is not None:
            console.success("Scans listed successfully.")
            columns = {
                "Id": "id",
                "Asset": "asset",
                "Created Time": "created_time",
                "Progress": "progress",
                "Risk Rating": "risk_rating",
            }
            title = f"Showing {len(scans)} Scans"
            data = [
                {
                    "id": str(s.id),
                    "asset": s.asset or _prepare_asset_str(s.id),
                    "created_time": str(s.created_time),
                    "progress": styles.style_progress(s.progress),
                    "risk_rating": styles.style_risk(_get_risk_rating(s.risk_rating)),
                }
                for s in scans
            ]

            console.table(columns=columns, data=data, title=title)
        else:
            console.error("Error fetching scan list.")


def _get_risk_rating(risk: Union[risk_rating.RiskRating, str, None]) -> str:
    """Get the risk rating string."""
    if risk is None:
        return "UNKNOWN"
    elif isinstance(risk, risk_rating.RiskRating):
        return risk.name
    else:
        return risk


def _prepare_asset_str(scan_id: int) -> str:
    """Prepare the asset string for the scan."""
    with models.Database() as session:
        assets = (
            session.query(models.Asset).filter(models.Asset.scan_id == scan_id).all()
        )
        if len(assets) == 0:
            return "N/A"

        asset_list = []
        for asset in assets:
            asset_type = asset.type
            if asset_type == "android_file":
                android_file_asset = (
                    session.query(models.AndroidFile).filter_by(id=asset.id).first()
                )
                asset_list.append(f"Android File: {android_file_asset.path}")
            elif asset_type == "ios_file":
                ios_file_asset = (
                    session.query(models.IosFile).filter_by(id=asset.id).first()
                )
                asset_list.append(f"IOS File: {ios_file_asset.path}")
            elif asset_type == "android_store":
                android_store_asset = (
                    session.query(models.AndroidStore).filter_by(id=asset.id).first()
                )
                asset_list.append(
                    f"Android Store: {android_store_asset.application_name or android_store_asset.package_name}"
                )
            elif asset_type == "ios_store":
                ios_store_asset = (
                    session.query(models.IosStore).filter_by(id=asset.id).first()
                )
                asset_list.append(
                    f"IOS Store: {ios_store_asset.application_name or ios_store_asset.bundle_id}"
                )
            elif asset_type == "network":
                ips = (
                    session.query(models.IPRange)
                    .filter_by(network_asset_id=asset.id)
                    .all()
                )
                networks_str = ", ".join(f"{ip.host}/{ip.mask or 32}" for ip in ips)
                asset_list.append(f"Network(s): {networks_str}")
            elif asset_type == "urls":
                urls = (
                    session.query(models.Link).filter_by(urls_asset_id=asset.id).all()
                )
                urls_str = ", ".join(url.url for url in urls)
                asset_list.append(f"Url(s): {urls_str}")
            elif asset_type == "domain_asset":
                domain = (
                    session.query(models.DomainName)
                    .filter_by(domain_asset_id=asset.id)
                    .first()
                )
                asset_list.append(f"Domain: {domain.name}")
            else:
                raise NotImplementedError(f"Asset type {asset.type} not implemented.")

        return ", ".join(asset_list)



"""create api key

Revision ID: cff0aef2c0f7
Revises: 4d13d49d324c
Create Date: 2024-05-29 17:08:31.735724

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "cff0aef2c0f7"
down_revision = "4d13d49d324c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "api_key",
        sa.Column("key", sa.String(length=150), nullable=False),
        sa.PrimaryKeyConstraint("key"),
        sa.UniqueConstraint("key"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("api_key")
    # ### end Alembic commands ###



"""Module for the command run inside the group scan.
This module takes care of preparing the selected runtime and the lists of provided agents, before starting a scan.
Example of usage:
    - oxo scan run --agent=agent1 --agent=agent2 --title=test_scan [asset] [options]."""

import io
import logging
from typing import List, Optional

import click
import httpx
import tenacity
from ruamel.yaml import error

from ostorlab import exceptions
from ostorlab.agent.schema import validator
from ostorlab.cli import console as cli_console
from ostorlab.cli import install_agent
from ostorlab.cli.scan import scan
from ostorlab.cli import types
from ostorlab.cli import agent_fetcher
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.utils import definitions as utils_definitions


console = cli_console.Console()

logger = logging.getLogger(__name__)

NUMBER_OF_RETRIES = 5
WAIT_BETWEEN_RETRIES = 5


@scan.group(invoke_without_command=True)
@click.option(
    "--agent",
    multiple=True,
    help="List agents keys (agent/<org>/<name> or @<org>/<name>). Org name can be omitted for defaults agent hosted "
    "by Ostorlab.",
    required=False,
    type=types.AgentKeyType(),
)
@click.option(
    "--arg",
    multiple=True,
    help="""Add an argument to an agent. The argument should be in the format: <name>:<value>.
     Example: --arg fast_mode:true
    """,
    type=types.AgentArgType(),
    required=False,
)
@click.option("--title", "-t", help="Scan title.")
@click.option(
    "--agent-group-definition",
    "-g",
    type=click.File("r"),
    help="Path to agents group definition file (yaml).",
    required=False,
)
@click.option(
    "--assets",
    "-a",
    type=click.File("r"),
    help="Path to target list definition file (yaml).",
    required=False,
)
@click.option(
    "--install", "-i", help="Install missing agents.", is_flag=True, required=False
)
@click.option(
    "--follow",
    help="Follow logs of provided list of agents given by their keys (agent/<org>/<name> or <org>/<name>). "
    "Org name can be omitted for defaults agents hosted by Ostorlab.",
    multiple=True,
    type=types.AgentKeyType(),
    default=[],
)
@click.option(
    "--no-follow",
    help="Start the scan without following the logs.",
    is_flag=True,
    default=False,
    required=False,
)
@click.option(
    "--no-tracker",
    help="Start the scan without tracker in charge of handling scan completion tracking.",
    is_flag=True,
    default=False,
    required=False,
)
@click.option(
    "--no-asset",
    help="Start the environment without injecting assets",
    is_flag=True,
    required=False,
)
@click.option(
    "--timeout",
    "-t",
    type=int,
    help="Timeout for the scan in seconds",
    required=False,
)
@click.option(
    "--init-sleep",
    type=int,
    help="Init sleep for tracker before checking the queues",
    required=False,
)
@click.pass_context
def run(
    ctx: click.core.Context,
    agent: List[str],
    arg: list[types.AgentArg],
    agent_group_definition: io.FileIO,
    assets: io.FileIO,
    title: str,
    install: bool,
    follow: List[str],
    no_follow: bool,
    no_tracker: bool,
    no_asset: bool,
    timeout: Optional[int] = None,
    init_sleep: Optional[int] = None,
) -> None:
    """Start a new scan on your assets.\n
    Example:\n
        - oxo scan run --agent=agent/ostorlab/nmap --agent=agent/google/tsunami --title=test_scan ip 8.8.8.8
    """
    if no_asset is True and ctx.invoked_subcommand is not None:
        console.error(
            f"Sub-command {ctx.invoked_subcommand} specified with --no-asset flag."
        )
        raise click.exceptions.Exit(2)
    if no_asset is False and ctx.invoked_subcommand is None and assets is None:
        console.error("Error: Missing command.")
        click.echo(ctx.get_help())
        raise click.exceptions.Exit(2)

    if agent:
        agents_settings: List[definitions.AgentSettings] = []
        for agent_key in agent:
            agents_settings.append(definitions.AgentSettings(key=agent_key))

        agent_group = definitions.AgentGroupDefinition(agents=agents_settings)
    elif agent_group_definition:
        try:
            agent_group = definitions.AgentGroupDefinition.from_yaml(
                agent_group_definition
            )
        except validator.ValidationError as e:
            console.error("Invalid agent group definition.")
            console.print(f"{e}")
            raise click.exceptions.Exit(2)
        except error.YAMLError as e:
            console.error("Agent group definition YAML parse error:")
            console.print(f"{e}")
            raise click.exceptions.Exit(2)
    else:
        raise click.ClickException("Missing agent list or agent group definition.")

    asset_group = None
    if assets is not None:
        try:
            asset_group = definitions.AssetsDefinition.from_yaml(assets)
        except validator.ValidationError as e:
            console.error(f"{e}")
            raise click.ClickException("Invalid asset Group Definition.") from e
    runtime_instance: runtime.Runtime = ctx.obj["runtime"]

    if no_tracker is True:
        runtime_instance.has_tracker = False

    if timeout is not None:
        runtime_instance.timeout = timeout

    if init_sleep is not None:
        runtime_instance.init_sleep = init_sleep

    # Prepare and set the list of agents to follow.
    agent_keys = [agent.key for agent in agent_group.agents]
    runtime_instance.follow = prepare_agents_to_follow(agent_keys, follow, no_follow)

    try:
        can_run_scan = runtime_instance.can_run(agent_group_definition=agent_group)
    except exceptions.OstorlabError as e:
        console.error(f"{e}")
        raise click.ClickException("Runtime encountered an error to run scan") from e

    if can_run_scan is True:
        ctx.obj["agent_group_definition"] = agent_group
        ctx.obj["title"] = title
        if install is True:
            try:
                _install_agents_with_retry(runtime_instance, agent_group)
            except httpx.HTTPError as e:
                console.error(f"Could not install the agents: {e}")
                raise click.exceptions.Exit(1)

        if arg is not None and len(arg) > 0:
            agent_group.agents = _add_cli_args_to_agent_settings(
                agent_group.agents, cli_args=arg
            )
        if ctx.invoked_subcommand is None:
            try:
                created_scan = runtime_instance.scan(
                    title=ctx.obj["title"],
                    agent_group_definition=ctx.obj["agent_group_definition"],
                    assets=asset_group.targets if asset_group is not None else None,
                )
                if created_scan is not None:
                    runtime_instance.link_agent_group_scan(created_scan, agent_group)
                    runtime_instance.link_assets_scan(
                        created_scan.id,
                        asset_group.targets if asset_group is not None else None,
                    )

            except exceptions.OstorlabError as e:
                console.error(f"An error was encountered while running the scan: {e}")
    else:
        raise click.ClickException(
            "The runtime does not support the provided agent list or group definition."
        )


@tenacity.retry(
    stop=tenacity.stop.stop_after_attempt(NUMBER_OF_RETRIES),
    wait=tenacity.wait.wait_fixed(WAIT_BETWEEN_RETRIES),
    retry=tenacity.retry_if_exception_type((httpx.HTTPError)),
    retry_error_callback=lambda retry_state: retry_state.outcome.result(),
)
def _install_agents_with_retry(
    runtime_instance: runtime.Runtime, agent_group: definitions.AgentGroupDefinition
) -> None:
    # Trigger both the runtime installation routine and install all the provided agents.
    runtime_instance.install()
    for ag in agent_group.agents:
        try:
            if ag.version is None:
                try:
                    agent_details = agent_fetcher.get_details(ag.key)
                    ag.version = agent_details["versions"]["versions"][0]["version"]
                except agent_fetcher.AgentDetailsNotFound:
                    logger.warning(
                        "agent %s not found on the store, skipping version fetch",
                        ag.key,
                    )
            install_agent.install(ag.key, ag.version)
        except agent_fetcher.AgentDetailsNotFound:
            console.warning(f"agent {ag.key} not found on the store")


def prepare_agents_to_follow(
    agent_keys: List[str], follow: List[str], no_follow: bool
) -> set[str]:
    """
    Prepares the list of agents to follow based on the provided list of agents to follow and unfollow.

    Args:
        agent_keys : List of agent keys.
        follow : List of agents to follow.
        no_follow : Flag to determine if we should log the agents.

    Returns:
        The list of agents to follow.
    """
    # If no_follow is True, we don't follow any agent.
    if no_follow is True:
        return set()

    # If follow is provided, we follow only the provided agents.
    if len(follow) > 0:
        return set(follow)

    # If follow is not provided, we follow all the agents.
    return set(agent_keys)


def _add_cli_args_to_agent_settings(
    agents_settings: list[definitions.AgentSettings], cli_args: list[types.AgentArg]
) -> list[definitions.AgentSettings]:
    """
    Adds CLI arguments to the agent settings if they are supported by the agent.

    Args:
        agents_settings : A list of agent settings.
        cli_args : A list of CLI Agent arguments.

    Returns:
         The updated list of agent settings.

    Raises:
        click.exceptions.Exit: If agent details are not found.
    """
    for agent_setting in agents_settings:
        try:
            agent_definition = agent_fetcher.get_definition(
                agent_key=agent_setting.key, version=agent_setting.version
            )
        except agent_fetcher.AgentDetailsNotFound:
            console.error(f"Agent {agent_setting.key} not found.")
            raise click.exceptions.Exit(2)

        for cli_argument in cli_args:
            for arg in agent_definition.args:
                if arg.get("name") == cli_argument.name:
                    try:
                        agent_setting.args.append(
                            utils_definitions.Arg.build(
                                name=cli_argument.name,
                                type=arg.get("type", "string"),
                                value=cli_argument.value,
                            )
                        )
                        break
                    except ValueError as e:
                        console.warning(
                            f"Could not set argument {cli_argument.name} to {cli_argument.value} because of "
                            f"the following error: {e}"
                        )
    return agents_settings



"""Agent List command."""

import logging
import re
from typing import Optional

import click
import docker

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent as agent_cli

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent_cli.command()
@click.argument("agent", required=True)
@click.option(
    "--agent-version-regex",
    "-r",
    help="Agent version matching regular expression.",
    required=False,
)
def delete(agent: str, agent_version_regex: Optional[str] = None) -> None:
    """CLI command to delete installed agents."""
    docker_client = docker.from_env()
    images = docker_client.images.list()
    agent_container_name = agent.replace("/", "_")
    deleted = False
    for im in images:
        # If any tag match the agent key tag, we delete all tags. This is because images from the store downloaded with
        # store URL, and will cause the image to remain if we delete only one tag.
        if any(t.split(":")[0] == agent_container_name for t in im.tags):
            for t in im.tags:
                agent_container_version = t.split(":")[1]
                if agent_version_regex is None or re.match(
                    agent_version_regex, agent_container_version
                ):
                    console.info(f"deleting container [bold red]{t}[/]")
                    docker_client.images.remove(t, force=True)
                    console.success(
                        f"container image [bold red]{t}[/] deleted successfully"
                    )
                    deleted = True

    if deleted is False:
        console.error(f"No agent matching [bold white]{agent}[/] was found")



"""Common definition types."""

import dataclasses
import json
from typing import Optional, Union, Any


@dataclasses.dataclass
class Arg:
    """Data class holding a definition of an argument for an agent.
    The actual type of value will be determined using the type attribute.
    The type could be one of the following: string, number, boolean, binary array, or object."""

    name: str
    type: str
    value: Optional[Union[bytes, int, float, str, bool]] = None
    description: Optional[str] = None

    @classmethod
    def build(
        cls,
        name: str,
        type: str,
        value: Optional[Union[bytes, int, float, str, bool]] = None,
        description: Optional[str] = None,
    ) -> "Arg":
        if isinstance(value, bytes):
            if type == "binary":
                value = value
            # When the value comes from a message received in the NATS.
            else:
                value = Arg.convert_str(value_str=value.decode(), target_type=type)

        # When the value comes from the CLI arguments using --arg.
        elif isinstance(value, str):
            value = Arg.convert_str(value_str=value, target_type=type)

        # When the value comes from the CLI with a YAML file for the group definition.
        else:
            # In this case, we don't need to parse the value since it will already be in the correct type.
            pass

        return cls(name, type, value, description)

    @staticmethod
    def convert_str(value_str: str, target_type: str) -> Any:
        """
        Convert a string value to the specified type.

        Args:
            target_type: The desired data type to convert the value to.

            value_str: The string representation of the value.

        Returns:
            Any: The converted value of the specified type.

        Raises:
            ValueError: If the target_type is not supported or If the value_str cannot be converted to the specified type.
        """
        if target_type == "string":
            return value_str
        elif target_type == "int":
            return int(value_str)
        elif target_type == "number":
            return float(value_str)
        elif target_type == "boolean":
            return value_str.lower() == "true"
        elif target_type == "array":
            try:
                return json.loads(value_str)
            except ValueError:
                return value_str.split(",")
        elif target_type == "object":
            return json.loads(value_str)
        else:
            raise ValueError(f"Unsupported argument type: {target_type}")


@dataclasses.dataclass
class PortMapping:
    """Data class defining a port mapping source to destination"""

    source_port: int
    destination_port: int


@dataclasses.dataclass
class ScannerState:
    """Current scanner state."""

    scanner_id: str
    scan_id: Optional[int]
    cpu_load: float
    memory_load: float
    total_cpu: Optional[int]
    total_memory: int
    hostname: str
    ip: str



"""Utils to handle IP operations."""

from typing import Optional
import socket


def get_ip() -> Optional[str]:
    """Returns the machine IP address."""
    return socket.gethostbyname(socket.gethostname())



"""create assets models

Revision ID: 736027d2cc2f
Revises: 23afcfd1d789
Create Date: 2024-06-03 10:33:26.311585

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "736027d2cc2f"
down_revision = "23afcfd1d789"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "asset",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(length=50), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_asset")),
    )
    op.create_table(
        "android_file",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("package_name", sa.String(length=255), nullable=True),
        sa.Column("path", sa.String(length=1024), nullable=True),
        sa.ForeignKeyConstraint(
            ["id"], ["asset.id"], name=op.f("fk_android_file_id_asset")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_android_file")),
    )
    op.create_table(
        "android_store",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("package_name", sa.String(length=255), nullable=True),
        sa.Column("application_name", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["id"], ["asset.id"], name=op.f("fk_android_store_id_asset")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_android_store")),
    )
    op.create_table(
        "ios_file",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("bundle_id", sa.String(length=255), nullable=True),
        sa.Column("path", sa.String(length=1024), nullable=True),
        sa.ForeignKeyConstraint(
            ["id"], ["asset.id"], name=op.f("fk_ios_file_id_asset")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_ios_file")),
    )
    op.create_table(
        "ios_store",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("bundle_id", sa.String(length=255), nullable=True),
        sa.Column("application_name", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["id"], ["asset.id"], name=op.f("fk_ios_store_id_asset")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_ios_store")),
    )
    op.create_table(
        "network",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("networks", sa.String(length=1024), nullable=True),
        sa.ForeignKeyConstraint(["id"], ["asset.id"], name=op.f("fk_network_id_asset")),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_network")),
    )
    op.create_table(
        "urls",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("links", sa.String(length=1024), nullable=True),
        sa.ForeignKeyConstraint(["id"], ["asset.id"], name=op.f("fk_urls_id_asset")),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_urls")),
    )
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.add_column(sa.Column("asset_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            batch_op.f("fk_scan_asset_id_asset"), "asset", ["asset_id"], ["id"]
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_scan_asset_id_asset"), type_="foreignkey"
        )
        batch_op.drop_column("asset_id")

    op.drop_table("urls")
    op.drop_table("network")
    op.drop_table("ios_store")
    op.drop_table("ios_file")
    op.drop_table("android_store")
    op.drop_table("android_file")
    op.drop_table("asset")
    # ### end Alembic commands ###



"""Link asset."""

import dataclasses

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.link")
class Link(asset.Asset):
    """Agent asset."""

    url: str
    method: str

    def __str__(self) -> str:
        return f"Link {self.url} with method {self.method}"

    @property
    def proto_field(self) -> str:
        return "link"



"""Asset of type .IPA (IOS App Store Package).
This module takes care of preparing a file of type .IPA before injecting it to the runtime instance.
"""

import io
import logging
from typing import Tuple, Optional

import click

from ostorlab.assets import ios_ipa as ios_ipa_asset
from ostorlab.cli.scan.run import run
from ostorlab.cli import console as cli_console
from ostorlab import exceptions

console = cli_console.Console()
logger = logging.getLogger(__name__)


@run.run.command()
@click.option("--file", type=click.File(mode="rb"), multiple=True, required=False)
@click.option("--url", multiple=True, required=False)
@click.pass_context
def ios_ipa(
    ctx: click.core.Context,
    file: Optional[Tuple[io.FileIO]] = (),
    url: Optional[Tuple[str]] = (),
) -> None:
    """Run scan for .IPA package file."""
    runtime = ctx.obj["runtime"]
    assets = []

    if url != () and file != ():
        console.error("Command accepts either path or source url of the ipa file.")
        raise click.exceptions.Exit(2)
    if url == () and file == ():
        console.error("Command missing either file path or source url of the ipa file.")
        raise click.exceptions.Exit(2)

    if file != ():
        for f in file:
            assets.append(ios_ipa_asset.IOSIpa(content=f.read(), path=str(f.name)))
    if url != ():
        for u in url:
            assets.append(ios_ipa_asset.IOSIpa(content_url=u))

    logger.debug("scanning assets %s", [str(asset) for asset in assets])
    try:
        created_scan = runtime.scan(
            title=ctx.obj["title"],
            agent_group_definition=ctx.obj["agent_group_definition"],
            assets=assets,
        )
        if created_scan is not None:
            runtime.link_agent_group_scan(
                created_scan, ctx.obj["agent_group_definition"]
            )
            runtime.link_assets_scan(created_scan.id, assets)

    except exceptions.OstorlabError as e:
        console.error(f"An error was encountered while running the scan: {e}")



"""Switch asset types to enum

Revision ID: 69c7d2f49a4e
Revises: b31a27c8584f
Create Date: 2024-06-26 18:21:23.093382

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "69c7d2f49a4e"
down_revision = "b31a27c8584f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("asset_type", schema=None) as batch_op:
        batch_op.alter_column(
            "type",
            existing_type=sa.VARCHAR(length=255),
            type_=sa.Enum(
                "ANDROID_FILE",
                "IOS_FILE",
                "ANDROID_STORE",
                "IOS_STORE",
                "LINK",
                "IP",
                "DOMAIN",
                "FILE",
                name="assettypeenum",
            ),
            existing_nullable=True,
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("asset_type", schema=None) as batch_op:
        batch_op.alter_column(
            "type",
            existing_type=sa.Enum(
                "ANDROID_FILE",
                "IOS_FILE",
                "ANDROID_STORE",
                "IOS_STORE",
                "LINK",
                "IP",
                "DOMAIN",
                "FILE",
                name="assettypeenum",
            ),
            type_=sa.VARCHAR(length=255),
            existing_nullable=True,
        )
    # ### end Alembic commands ###



"""Check if requirements for running docker are satisfied."""

import platform
import sys

import docker
import tenacity
from docker import errors

from ostorlab import exceptions

_SUPPORTED_ARCH_TYPES = ["x86_64", "AMD64", "arm64", "aarch64"]
RETRY_ATTEMPTS = 10
WAIT_TIME = 2

# The architecture is checked with a return value that's based on the kernel implementation of the uname(2)
# system call. So it might be necessary to handle the same arch with various strings e.g. linux returns x86_64
# or AMD64 on windows.


def is_docker_installed() -> bool:
    """Checks if docker is installed

    Returns:
        True if docker is installed, else False
    """
    try:
        _ = docker.from_env()
    except errors.DockerException as e:
        if "ConnectionRefusedError" in str(e):
            return False
    return True


def is_sys_arch_supported() -> bool:
    """Checks if the systems cpu architecture is supported

    Returns:
        True if the architecture is supported, else False
    """
    if platform.machine() not in _SUPPORTED_ARCH_TYPES:
        return False
    return True


def is_user_permitted() -> bool:
    """Check if the user got permissions to run docker.

    Returns:
        True if user has permission to run docker, else False
    """
    try:
        _ = docker.from_env()
    except errors.DockerException as e:
        if "PermissionError" in str(e):
            return False
    return True


def is_docker_working() -> bool:
    """Last hope check to see if docker works without being able to give an intelligible recommendation.

    Returns:
        True if user has permission to run docker, else False
    """
    if sys.platform == "win32":
        import pywintypes

        try:
            client = docker.from_env()
            client.ping()
        except errors.DockerException:
            return False
        except pywintypes.error:
            return False
    else:
        try:
            client = docker.from_env()
            client.ping()
        except errors.DockerException:
            return False
    return True


def is_swarm_initialized() -> bool:
    """Checks if docker swarm is initialized.

    Returns:
        True if docker swarm is initialized, else False
    """
    if is_user_permitted():
        docker_client = docker.from_env()
        if docker_client.swarm.id is None:
            return False
        else:
            return True
    else:
        return False


def init_swarm() -> None:
    """Initializes Docker Swarm.

    This function attempts to initialize Docker Swarm. If the initialization fails,
    it retries 10 times with a 2-second  delay between each attempt.
    If it still fails after 10 attempts, it raises an OstorlabError.

    Raises:
        OstorlabError: If the user does not have permission to run Docker,
        or if the initialization fails after 10 attempts.
    """
    if is_user_permitted() is False:
        raise errors.DockerException("User does not have permission to run docker.")
    try:
        _init_swarm()
    except errors.DockerException as e:
        raise exceptions.OstorlabError("Error while initializing swarm.") from e


@tenacity.retry(
    stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS),
    wait=tenacity.wait_fixed(WAIT_TIME),
    retry=tenacity.retry_if_exception_type(errors.DockerException),
    reraise=True,
)
def _init_swarm() -> None:
    """Initialize docker swarm"""
    docker_client = docker.from_env()
    docker_client.swarm.init()



"""File asset."""

import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.file")
class File(asset.Asset):
    """File target asset."""

    content: Optional[bytes] = None
    path: Optional[str] = None
    content_url: Optional[str] = None

    def __str__(self) -> str:
        str_representation = "File"
        if self.path is not None:
            str_representation = f"{str_representation}:{self.path}"
        if self.content_url is not None:
            str_representation = f"{str_representation}:{self.content_url}"

        return str_representation

    @property
    def proto_field(self) -> str:
        return "file"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/capture/stack_trace/stack_trace.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nEostorlab/agent/message/proto/v3/capture/stack_trace/stack_trace.proto\x12\x33ostorlab.agent.message.proto.v3.capture.stack_trace\"0\n\x03\x41rg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x0c\n\x04type\x18\x03 \x01(\t\"\xc9\x01\n\x05\x46rame\x12\x11\n\tfile_path\x18\x01 \x01(\t\x12\x15\n\rfunction_name\x18\x02 \x01(\t\x12\x12\n\nclass_name\x18\x03 \x01(\t\x12\x14\n\x0cpackage_name\x18\x04 \x01(\t\x12\x13\n\x0breturn_type\x18\x05 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x06 \x01(\x04\x12\x46\n\x04\x61rgs\x18\x07 \x03(\x0b\x32\x38.ostorlab.agent.message.proto.v3.capture.stack_trace.Arg\"2\n\x05Input\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x0c\n\x04type\x18\x03 \x01(\t\"\xc2\x01\n\x07Message\x12J\n\x06\x66rames\x18\x01 \x03(\x0b\x32:.ostorlab.agent.message.proto.v3.capture.stack_trace.Frame\x12\x11\n\tthread_id\x18\x02 \x01(\t\x12\x0c\n\x04time\x18\x03 \x01(\x02\x12J\n\x06inputs\x18\x04 \x03(\x0b\x32:.ostorlab.agent.message.proto.v3.capture.stack_trace.Input')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.capture.stack_trace.stack_trace_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ARG._serialized_start=126
  _ARG._serialized_end=174
  _FRAME._serialized_start=177
  _FRAME._serialized_end=378
  _INPUT._serialized_start=380
  _INPUT._serialized_end=430
  _MESSAGE._serialized_start=433
  _MESSAGE._serialized_end=627
# @@protoc_insertion_point(module_scope)



"""Stream logs of a service from a thread."""

import threading
import time

import docker.errors
from docker.models import services as docker_service

from ostorlab.cli import console as cli_console


console = cli_console.Console()

COLOR_POOL = [
    "dodger_blue2",
    "deep_sky_blue3",
    "deep_sky_blue2",
    "cyan3",
    "spring_green2",
    "spring_green2",
    "grey37",
    "chartreuse4",
    "cornflower_blue",
    "chartreuse3",
    "steel_blue1",
    "dark_red",
    "plum4",
]


class _ServiceLogStream:
    """Stream logs from a specific docker service."""

    def __init__(self, service: docker_service.Service, color: str) -> None:
        self._service: docker_service.Service = service
        self._color: str = color
        self._stop_event = threading.Event()
        self._thread: threading.Thread | None = None

    def start(self) -> None:
        """Starts the service log stream, in the background."""
        self._thread = threading.Thread(target=self._stream, daemon=True)
        self._thread.start()

    def _stream(self) -> None:
        logs = self._service.logs(details=False, follow=True, stdout=True, stderr=True)
        name = self._service.name
        for line in logs:
            if self._stop_event.is_set():
                break
            console.info(f"[{self._color} bold]{name}:[/] {line[:-1].decode()}")

    def stop(self) -> None:
        """Stop the log stream."""
        if self._thread is None:
            return
        self._stop_event.set()
        # thread may check the stop event late when the thread calling LogStream.wait has already exited
        # so we join the thread with a timeout.
        self._thread.join(timeout=1)


class LogStream:
    """Docker service log streamer."""

    def __init__(self, docker_client: docker.DockerClient) -> None:
        self._color_map: dict[str, str] = {}
        self._log_streams: dict[str, _ServiceLogStream] = {}
        self._docker_client = docker_client

    def stream(self, service: docker_service.Service) -> None:
        """
        Stream logs of a service without blocking.

        Args:
            service: Docker service.
        """
        color = self._select_color(service)
        service_id = service.id

        if service_id in self._log_streams:
            return

        log_streamer = _ServiceLogStream(service, color=color)
        self._log_streams[service_id] = log_streamer
        log_streamer.start()

    def wait(self) -> None:
        """Wait for all streams to finish."""
        # The logs generator returned by the docker sdk does not return when the service is removed.
        # So we need to loop and poll each service to stop the stream manually. Thank you Docker.
        while len(self._log_streams) > 0:
            for service_id, log_stream in list(self._log_streams.items()):
                if self._service_exists(service_id) is False:
                    log_stream.stop()
                    del self._log_streams[service_id]
            time.sleep(1)

    def _select_color(self, service):
        """Select color for console output of service."""
        if service.name in self._color_map:
            color = self._color_map[service.name]
        else:
            # To avoid running out of colors, we insert them back.
            color = COLOR_POOL.pop()
            COLOR_POOL.insert(0, color)
        return color

    def _service_exists(self, service_id) -> bool:
        try:
            self._docker_client.services.get(service_id)
            return True
        except docker.errors.NotFound:
            return False



"""Vulnz dump command."""

import logging

import click
import sqlalchemy
from ostorlab.apis.runners import runner

from ostorlab.cli import console as cli_console
from ostorlab.cli import dumpers
from ostorlab.cli.vulnz import vulnz

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name="dump")
@click.option("--scan-id", "-s", "scan_id", help="Id of the scan.", required=True)
@click.option(
    "--output", "-o", help="Output file path", required=False, default="./output"
)
@click.option(
    "--format",
    "-f",
    "output_format",
    help="Output format",
    required=False,
    type=click.Choice(["jsonl", "csv"]),
    default="jsonl",
)
@click.pass_context
def dump(ctx, scan_id: int, output: str, output_format: str) -> None:
    """Dump found vulnerabilities of a scan in a specific format."""
    try:
        runtime_instance = ctx.obj["runtime"]
        if output_format == "csv":
            dumper = dumpers.VulnzCsvDumper(output)
        else:
            dumper = dumpers.VulnzJsonDumper(output)
        runtime_instance.dump_vulnz(scan_id=scan_id, dumper=dumper)
    except FileNotFoundError as e:
        console.error(f"No such file or directory: {output}")
        raise click.exceptions.Exit(2) from e
    except (sqlalchemy.exc.OperationalError, runner.Error):
        console.error(f"scan with id {scan_id} does not exist.")



"""Ios testflight target asset"""

import dataclasses

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.store.ios_testflight")
class IOSTestflight(asset.Asset):
    """iOS testflight target asset."""

    application_url: str

    def __str__(self) -> str:
        return f"iOS Testflight ({self.application_url})"

    @property
    def proto_field(self) -> str:
        return "ios_testflight"



"""Auth revoke command."""

import logging

from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner
from ostorlab.apis import revoke_api_key
from ostorlab import configuration_manager
from ostorlab.cli.auth import auth
from ostorlab.cli import console as cli_console

logger = logging.getLogger(__name__)
console = cli_console.Console()


@auth.auth.command()
def revoke():
    """Use this to revoke your API key."""

    config_manager = configuration_manager.ConfigurationManager()

    try:
        api_key_id = config_manager.api_key_id
        api_runner = authenticated_runner.AuthenticatedAPIRunner()

        with console.status("Revoking API key"):
            try:
                api_runner.execute(revoke_api_key.RevokeAPIKeyAPIRequest(api_key_id))
                api_runner.unauthenticate()
                config_manager.delete_api_data()
                console.success("API key revoked")
            except runner.ResponseError:
                console.error("Could not revoke your API key.")

    except authenticated_runner.AuthenticationError:
        api_runner.unauthenticate()
        console.success("Your API key has already been revoked.")



"""OpenTelemetry Mixin.

The mixin overrides the behaviour of the main methods of the agent, mainly emit and process to send traces,
metadata, metrics and exceptions.
"""

import base64
import io
import json
import logging
import os
import sys
import tempfile
import uuid
from typing import Any, Dict, Optional
from urllib import parse

from opentelemetry import trace
from opentelemetry.exporter import cloud_trace
from opentelemetry.exporter.jaeger import thrift as jaeger
from opentelemetry.sdk import resources
from opentelemetry.sdk import trace as trace_provider
from opentelemetry.sdk.trace import export as sdk_export

from ostorlab.agent import definitions as agent_definitions
from ostorlab.agent.message import message as agent_message
from ostorlab.runtimes import definitions as runtime_definitions
from ostorlab.utils import dictionary_minifier

logger = logging.getLogger(__name__)


class TraceExporter:
    """Class responsible for preparing the respective OpenTelemetry Span Exporter for an input tracing url."""

    def __init__(self, tracing_collector_url: str):
        self._tracing_collector_url = tracing_collector_url
        # specialized fields for the different collectors.
        self._file: Optional[io.IOBase] = None

    def close(self):
        if self._file is not None:
            self._file.close()

    def get_trace_exporter(self) -> sdk_export.SpanExporter:
        """
        Returns a span exporter instance depending on the OpenTelemtry collector url argument.
        The urls are customized to respect the following format:
            name_of_the_tracing_tool:hostname:port
            eg: jaeger:jaeger-host:8631
            for gcp the format is:
            name_of_the_tracing_tool://project_id/service_account_json_base64
            eg: gcp://project_1/service_account_json_base64
        """
        parsed_url = parse.urlparse(self._tracing_collector_url)
        scheme = parsed_url.scheme
        if scheme == "jaeger":
            return self._get_jaeger_exporter(parsed_url)
        elif scheme == "file":
            return self._get_file_exporter(parsed_url)
        elif scheme == "gcp":
            return self._get_gcp_exporter(parsed_url)
        else:
            raise NotImplementedError(f"Invalid tracer type {scheme}")

    def _get_file_exporter(self, parsed_url):
        file_path = parsed_url.path
        self._file = open(file_path, "w", encoding="utf-8")
        file_exporter = sdk_export.ConsoleSpanExporter(out=self._file)
        logger.info("Configuring file exporter..")
        return file_exporter

    def _get_jaeger_exporter(self, parsed_url):
        netloc = parsed_url.netloc
        hostname, port = netloc.split(":")[0], int(netloc.split(":")[1])
        jaeger_exporter = jaeger.JaegerExporter(
            agent_host_name=hostname,
            agent_port=port,
            udp_split_oversized_batches=True,
        )
        logger.info("Configuring jaeger exporter..")
        return jaeger_exporter

    def _get_gcp_exporter(self, parsed_url) -> cloud_trace.CloudTraceSpanExporter:
        """
        Returns a CloudTraceSpan exporter instance.
        The urls should respect the following format:
            project_id/service_account_json_base64_value
        """
        project_id = parsed_url.netloc
        # write service account key to temp file
        with tempfile.NamedTemporaryFile(
            mode="wb", delete=False
        ) as service_account_key_temp_file:
            service_account_json_content = base64.b64decode(
                parsed_url.path[1:].encode()
            )
            service_account_key_temp_file.write(service_account_json_content)
            # the env variable GOOGLE_APPLICATION_CREDENTIALS points to a file defining the service account credentials
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = (
                service_account_key_temp_file.name
            )
        return cloud_trace.CloudTraceSpanExporter(project_id=project_id)


class OpenTelemetryMixin:
    """OpenTelemetryMixin to send telemetry data of the agent behaviour. The mixin enables tracking information
    like when was a message processed or emitted, using which selector,
    how much time it took to serialize or deserialize a message, and report exceptions when they occur..etc.
    """

    _tracer: Optional[trace.Tracer] = None

    def __init__(
        self,
        agent_definition: agent_definitions.AgentDefinition,
        agent_settings: runtime_definitions.AgentSettings,
    ) -> None:
        """Initializes the mixin from the agent settings.

        Args:
            agent_settings: Agent runtime settings.
        """
        super().__init__(
            agent_definition=agent_definition, agent_settings=agent_settings
        )
        self._agent_settings = agent_settings
        if (
            self._agent_settings.tracing_collector_url is not None
            and self._agent_settings.tracing_collector_url != ""
        ):
            self._tracing_collector_url = self._agent_settings.tracing_collector_url
        else:
            self._tracing_collector_url = None

        if self._tracing_collector_url is not None:
            self._exporter = TraceExporter(self._tracing_collector_url)
            provider = trace_provider.TracerProvider(
                resource=resources.Resource.create(
                    {resources.SERVICE_NAME: agent_settings.key}
                )
            )
            trace.set_tracer_provider(provider)

            # NOTE: We have experienced crashes with BatchSpan processor on python version 3.9. The error is linked to
            # threading in the Python runtime. This warrants more investigation, this is a temporary fix until the
            # root cause is clearly identified.
            if sys.version_info >= (3, 10):
                self._span_processor = sdk_export.BatchSpanProcessor(
                    self._exporter.get_trace_exporter()
                )
            else:
                self._span_processor = sdk_export.SimpleSpanProcessor(
                    self._exporter.get_trace_exporter()
                )
            trace.get_tracer_provider().add_span_processor(self._span_processor)
            self._tracer = trace.get_tracer(__name__)

    @property
    def tracer(self) -> trace.Tracer:
        """Tracer object to report traces to the configured opentelemetry collector.
        To use:

        ```python
        ...
        with self.tracer.start_as_current_span('<action>') as process_msg_span:
                process_msg_span.set_attribute('<attr1>', <value1>)
                ...
        ```
        """
        return self._tracer

    def force_flush_file_exporter(self) -> None:
        """Ensures persistence of the span details in the file for the case of the file Span exporters."""
        self._span_processor.force_flush()

    def _stringify_bytes_values(self, value: bytes):
        """Method that will be used as a handler to json dump the message dictionary values."""
        if isinstance(value, bytes):
            return value.decode(errors="replace")
        else:
            return value

    def process_message(self, selector: str, message: bytes) -> None:
        """Overridden agent process message method to add OpenTelemetry traces.
        Processes raw message received from BS.

        Args:
            selector: destination selector with full path, including UUID set by default.
            message: raw bytes message.

        Returns:
            None
        """
        if self._tracing_collector_url is not None:
            logger.debug("recording process message trace.")

            selector_split = selector.split(".")
            message_id = selector_split[-1]

            message_id_split = message_id.split("-")
            if len(message_id_split) > 5:
                # message is passing the message id, trace id and span id.
                trace_uuid = int(message_id_split[5])
                span_uuid = int(message_id_split[6])
                parent_span_context = trace.SpanContext(
                    trace_id=trace_uuid,
                    span_id=span_uuid,
                    is_remote=True,
                    trace_flags=trace.TraceFlags(0x01),
                )
                context = trace.set_span_in_context(
                    trace.NonRecordingSpan(parent_span_context)
                )
            else:
                context = {}

            raw_selector = ".".join(selector_split[:-1])
            control_message = agent_message.Message.from_raw("v3.control", message)

            data = agent_message.Message.from_raw(
                raw_selector, control_message.data.get("message")
            ).data
            minified_msg_data = dictionary_minifier.minify_dict(
                data, dictionary_minifier.truncate_str
            )
            stringified_msg_data = json.dumps(
                minified_msg_data, default=self._stringify_bytes_values
            )

            with self.tracer.start_as_current_span(
                "process_message", context=context
            ) as process_msg_span:
                process_msg_span.set_attribute("agent.name", self.name)
                process_msg_span.set_attribute("message.selector", raw_selector)
                process_msg_span.set_attribute("message.data", stringified_msg_data)

                super().process_message(selector, message)
        else:
            super().process_message(selector, message)

    def emit(
        self, selector: str, data: Dict[str, Any], message_id: Optional[str] = None
    ) -> None:
        """Overridden emit method of the agent to add OpenTelemetry traces.
        Sends a message to all listening agents on the specified selector.

        Args:
            selector: target selector.
            data: message data to be serialized.
            message_id: An id that will be added to the tail of the message.
        Raises:
            NonListedMessageSelectorError: when selector is not part of listed out selectors.

        Returns:
            None
        """
        if self._tracing_collector_url is not None:
            with self.tracer.start_as_current_span("emit_message") as emit_span:
                logger.debug("recording emit message trace..")

                trace_id = emit_span.get_span_context().trace_id
                span_id = emit_span.get_span_context().span_id
                super().emit(
                    selector, data, f"{message_id or uuid.uuid4()}-{trace_id}-{span_id}"
                )

                emit_span.set_attribute("agent.name", self.name)
                emit_span.set_attribute("message.selector", selector)
                minified_msg_data = dictionary_minifier.minify_dict(
                    data, dictionary_minifier.truncate_str
                )
                emit_span.set_attribute("message.data", json.dumps(minified_msg_data))
        else:
            super().emit(selector, data)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/ip/v4/port/service/service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nFostorlab/agent/message/proto/v3/asset/ip/v4/port/service/service.proto\x12\x38ostorlab.agent.message.proto.v3.asset.ip.v4.port.service\"\x89\x01\n\x07Message\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04mask\x18\x02 \x01(\t\x12\x12\n\x07version\x18\x03 \x01(\x05:\x01\x34\x12\x0c\n\x04port\x18\x05 \x01(\r\x12\x10\n\x08protocol\x18\x06 \x01(\t\x12\r\n\x05state\x18\x07 \x01(\t\x12\x0f\n\x07service\x18\x08 \x01(\t\x12\x0e\n\x06\x62\x61nner\x18\t \x01(\t')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.asset.ip.v4.port.service.service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=133
  _MESSAGE._serialized_end=270
# @@protoc_insertion_point(module_scope)



"""Mixin to enable distributed storage for a group of agent replicas or for a single agent that needs reliable storage.

A typical use-case is ensuring that a message value is processed only once in case multiple other agent are producing
duplicates or similar ones.

Typical usage:

```python
    status_agent = agent_persist_mixin.AgentPersistMixin(agent_settings)
    status_agent.set_add('crawler_agent_asset_dna', dna)
    is_new = not status_agent.set_is_member()
```
"""

import ipaddress
import logging
from typing import Dict, Set, Callable, Optional, Union

import redis

from ostorlab.runtimes import definitions as runtime_definitions

logger = logging.getLogger(__name__)


class AgentPersistMixin:
    """Agent mixin to persist data. The mixin enables distributed storage of a group of agent replicas or for a single
    agent that needs reliable storage."""

    def __init__(self, agent_settings: runtime_definitions.AgentSettings) -> None:
        """Initializes the mixin from the agent settings.

        Args:
            agent_settings: Agent runtime settings.
        """
        if agent_settings.redis_url is None:
            raise ValueError("agent settings is missing redis url")
        self._redis_client = redis.Redis.from_url(agent_settings.redis_url)

    def set_add(self, key: Union[bytes, str], *value: Union[bytes, str]) -> bool:
        """Helper function that takes care of reporting if the specified DNA has been tested in the past, or mark it
        as tested.
        The method can be used to sync multiple agents that may encounter the same test input but need to test it
        only once.
        Storage is shared between agents, ensure the keys used are unique.

        Args:
            key: Set key.
            value: values to add to set.

        Returns:
            True if it is a new member, False otherwise.
        """
        return bool(self._redis_client.sadd(key, *value))

    def set_is_member(self, key: Union[bytes, str], value: Union[bytes, str]) -> bool:
        """Indicates whether value is member of the set identified by key.

        Args:
            key: Set key.
            value: The value to check.

        Returns:
            True if it is a member, False otherwise.
        """
        return self._redis_client.sismember(key, value) == 1

    def set_len(self, key: Union[bytes, str]) -> int:
        """Helper function that returns the set cardinality (number of elements) of the set stored at key.
        The method can be used to sync multiple agents that may receive test inputs but need to test
        less than X test inputs.
        Storage is shared between agents, ensure the keys used are unique.

        Args:
            key: Set key.

        Returns:
            The cardinality (number of elements) of the set, or 0 if key does not exist.
        """
        return self._redis_client.scard(key)

    def set_members(self, key: Union[bytes, str]) -> Set[bytes]:
        """Helper function that returns all the members of the set value stored at key.

        Args:
            key: Set key.

        Returns:
             The value of key, or empty set when key does not exist.
        """
        return self._redis_client.smembers(key)

    def add(self, key: Union[bytes, str], value: bytes) -> Optional[bool]:
        """Helper function that Set key to hold the string value.

        Args:
            key: Key for string value.
            value: Bytes value.

        Returns:
            Status of the set command.
        """
        return self._redis_client.set(key, value)

    def get(self, key: Union[bytes, str]) -> Optional[bytes]:
        """Get the value of key. If the key does not exist None is returned.

        Args:
            key: Key for string value.

        Returns:
            Value of key, or None when key does not exist.
        """
        return self._redis_client.get(key)

    def exists(self, key: Union[bytes, str]) -> bool:
        """Helper function that checks if a given key exists in the Redis database.

        Args:
            key: The key to check for existence.

        Returns:
            bool: True if the key exists, False otherwise.
        """

        return self._redis_client.exists(key) == 1

    def hash_add(
        self,
        hash_name: Union[bytes, str],
        mapping: Dict[Union[bytes, str], Union[bytes, str]],
    ) -> bool:
        """Set mapping within hash hash_name. If hash_name does not exist a new hash is created.
        If key exists, value is overridden.

        Args:
            hash_name: Name of the hash.
            mapping: Dict of key/value pairs that will be added to hash_name

        Returns:
            True if the key does not exist, False otherwise
        """
        return bool(self._redis_client.hset(name=hash_name, mapping=mapping))

    def hash_exists(self, hash_name: Union[bytes, str], key: Union[bytes, str]) -> bool:
        """Returns a boolean indicating if key exists within hash hash_name.

        Args:
            hash_name: Name of the hash.
            key: Key in the hash.

        Returns:
            True if the key exists in the hash.
        """
        return self._redis_client.hexists(hash_name, key)

    def hash_get(
        self, hash_name: Union[bytes, str], key: Union[bytes, str]
    ) -> Optional[bytes]:
        """Return the value of key within the hash hash_name.

        Args:
            hash_name: Name of the hash.
            key: Key in the hash.

        Returns:
            Value of the key stored in hash_name, None if the key does not exist.
        """
        return self._redis_client.hget(hash_name, key)

    def hash_get_all(self, hash_name: Union[bytes, str]) -> Dict:
        """Returns a dict of the hash’s name/value pairs.

        Args:
            hash_name: Name of the hash.

        Returns:
            Dict of the hash’s name/value pairs.
        """
        return self._redis_client.hgetall(hash_name)

    def delete(self, key: Union[bytes, str]) -> bool:
        """Delete a specific key.

        Args:
            key: Key of any type.

        Returns:
            True if the key is delete, False otherwise.
        """
        return bool(self._redis_client.delete(key))

    def value_type(self, key: Union[bytes, str]) -> str:
        """Return a string representation of the type of the value stored at key.

        Args:
            key: Key in redis.

        Returns:
            String representation of the type of value stored at key. eg: set, string.
        """
        return self._redis_client.type(key).decode()

    def add_ip_network(
        self,
        key: Union[bytes, str],
        ip_range: Union[ipaddress.IPv6Network, ipaddress.IPv4Network],
        value: Optional[
            Callable[
                [Union[ipaddress.IPv6Network, ipaddress.IPv4Network]], Union[bytes, str]
            ]
        ] = None,
    ) -> bool:
        """
        Returns True if a network have never been persisted before, else it's returns False
        this method takes
        Args:
            key: Unique key for the set or a callable that gets the ip network getting check and returns the key.
            ip_range: IPv4Network or IPv4Network network to persist
            value: Callable to format the IP network. This is helpful to add extra data to the range, like for instance
             the port or service in case many needs to be tested separately.

        Returns:
            returns True if network is added and False if the network or one of it's super nets already exits.
        """
        ip_network = ip_range
        while ip_network.prefixlen != 0:
            if value is not None:
                member_value = value(ip_network)
            else:
                member_value = ip_network.exploded
            if self.set_is_member(key, member_value) is True:
                return False
            ip_network = ip_network.supernet()

        if value is not None:
            member_value = value(ip_range)
        else:
            member_value = ip_range.exploded
        return self.set_add(key, member_value)

    def ip_network_exists(
        self,
        key: Union[bytes, str],
        ip_range: Union[ipaddress.IPv6Network, ipaddress.IPv4Network],
        value: Optional[
            Callable[
                [Union[ipaddress.IPv6Network, ipaddress.IPv4Network]], Union[bytes, str]
            ]
        ] = None,
    ) -> bool:
        """
        Returns False if a network have never been persisted before, else it returns True.

        this method takes
        Args:
            key: key for the set
            ip_range: IPv4Network or IPv4Network network to check
            value: Callable to format the IP network. This is helpful to add extra data to the range, like for instance
             the port or service in case many needs to be tested separately.

        Returns:
            returns False if network is never been persisted (diff to add_ip_network shouldn't add it)
            and True if the network or one of it's super nets already exits.
        """
        ip_network = ip_range
        while ip_network.prefixlen != 0:
            if value is not None:
                member_value = value(ip_network)
            else:
                member_value = ip_network.exploded
            if self.set_is_member(key, member_value) is True:
                return True
            ip_network = ip_network.supernet()

        return False



"""Asset of type link.
This module takes care of preparing a link before calling the create web scan API.
"""

import io
import json

import click
import itertools

from typing import List, Optional

from ostorlab.cli.ci_scan.run import run
from ostorlab.apis import scan_create as scan_create_api
from ostorlab.apis import test_credentials_create as test_credentials_create_api
from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner as base_runner


def _prepare_test_credentials(
    ctx: click.core.Context,
) -> List[test_credentials_create_api.TestCredential]:
    test_credentials_login = ctx.obj["test_credentials"]["test_credentials_login"]
    test_credentials_password = ctx.obj["test_credentials"]["test_credentials_password"]
    test_credentials_url = ctx.obj["test_credentials"]["test_credentials_url"]
    test_credentials_role = ctx.obj["test_credentials"]["test_credentials_role"]
    test_credentials_name = ctx.obj["test_credentials"]["test_credentials_name"]
    test_credentials_value = ctx.obj["test_credentials"]["test_credentials_value"]

    credentials = []
    for login, password, role, url in itertools.zip_longest(
        test_credentials_login,
        test_credentials_password,
        test_credentials_role,
        test_credentials_url,
    ):
        credentials.append(
            test_credentials_create_api.TestCredentialLogin(
                login=login, password=password, role=role, url=url
            )
        )

    if test_credentials_name:
        credentials.append(
            test_credentials_create_api.TestCredentialCustom(
                values=dict(
                    itertools.zip_longest(test_credentials_name, test_credentials_value)
                )
            )
        )

    return credentials


def run_link_scan(ctx: click.core.Context, url: List[str]) -> None:
    """Create scan for links."""
    ci_logger = ctx.obj["ci_logger"]
    if ctx.obj.get("api_key") is not None:
        scan_profile = ctx.obj["scan_profile"]
        title = ctx.obj["title"]
        break_on_risk_rating = ctx.obj["break_on_risk_rating"]
        max_wait_minutes = ctx.obj["max_wait_minutes"]
        sboms = ctx.obj["sboms"]
        api_schema = ctx.obj["api_schema"]
        filtered_url_regexes = ctx.obj["filtered_url_regexes"]
        proxy = ctx.obj["proxy"]
        qps = ctx.obj["qps"]
        ui_prompt_ids = ctx.obj.get("ui_prompt_ids") or []
        ui_prompt_names = ctx.obj.get("ui_prompt_names") or []
        ui_prompt_actions = ctx.obj.get("ui_prompt_actions") or []
        runner = authenticated_runner.AuthenticatedAPIRunner(
            api_key=ctx.obj.get("api_key")
        )
        try:
            test_credentials = _prepare_test_credentials(ctx)
            if test_credentials:
                credential_ids = _create_test_credentials(test_credentials, runner)
                ci_logger.info(
                    f"Created test credentials {', '.join(str(t) for t in test_credentials)} successfully"
                )
            else:
                credential_ids = []

            ci_logger.info(
                f"creating Web scan `{title}` with profile `{scan_profile}`."
            )

            ui_automation_rule_ids: List[int] = []

            if len(ui_prompt_ids) > 0:
                ui_automation_rule_ids.extend(ui_prompt_ids)
                ci_logger.info(f"Using existing UI prompts with IDs: {ui_prompt_ids}")

            if len(ui_prompt_names) > 0 and len(ui_prompt_actions) > 0:
                ui_prompts_json = [
                    {"name": name, "code": action}
                    for name, action in zip(ui_prompt_names, ui_prompt_actions)
                ]
                try:
                    ci_logger.info("Creating UI prompts...")
                    prompts_result = runner.execute(
                        scan_create_api.CreateUIPromptsAPIRequest(
                            ui_prompts=ui_prompts_json
                        )
                    )
                    created_prompt_ids = [
                        int(prompt["id"])
                        for prompt in prompts_result["data"]["createUiPrompts"][
                            "uiPrompts"
                        ]
                    ]
                    ui_automation_rule_ids.extend(created_prompt_ids)
                    ci_logger.info(f"Created UI prompts with IDs: {created_prompt_ids}")
                except (json.JSONDecodeError, KeyError) as e:
                    ci_logger.error(
                        f"Invalid UI prompts format: {e}. Continuing without UI prompts."
                    )

            scan_id = _create_scan(
                title=title,
                scan_profile=scan_profile,
                urls=url,
                credential_ids=credential_ids,
                runner=runner,
                sboms=sboms,
                api_schema=api_schema,
                filtered_url_regexes=filtered_url_regexes,
                proxy=proxy,
                qps=qps,
                ui_automation_rule_ids=ui_automation_rule_ids,
            )

            ci_logger.output(name="scan_id", value=scan_id)
            ci_logger.info(f"Scan created with id {scan_id}.")

            if break_on_risk_rating is not None and break_on_risk_rating != "":
                run.apply_break_scan_risk_rating(
                    break_on_risk_rating, scan_id, max_wait_minutes, runner, ci_logger
                )

        except base_runner.ResponseError as e:
            ci_logger.error(f"Could not start the scan. {e}")
            raise click.exceptions.Exit(2) from e
        except TimeoutError:
            ci_logger.error("The scan is still running.")
            raise click.exceptions.Exit(2) from None
    else:
        ci_logger.error("API key not not provided.")
        raise click.exceptions.Exit(2) from None


def _create_scan(
    credential_ids: List[int],
    runner: authenticated_runner.AuthenticatedAPIRunner,
    title: str,
    urls: List[str],
    scan_profile: str,
    sboms: Optional[list[io.FileIO]] = None,
    api_schema: Optional[io.FileIO] = None,
    proxy: Optional[str] = None,
    qps: Optional[int] = None,
    filtered_url_regexes: Optional[List[str]] = None,
    test_credential_ids: Optional[List[int]] = None,
    ui_automation_rule_ids: List[int] = (),
):
    scan_result = runner.execute(
        scan_create_api.CreateWebScanAPIRequest(
            title=title,
            urls=urls,
            scan_profile=scan_profile,
            test_credential_ids=credential_ids,
            sboms=sboms,
            api_schema=api_schema,
            filtered_url_regexes=filtered_url_regexes,
            proxy=proxy,
            qps=qps,
            ui_automation_rule_ids=ui_automation_rule_ids,
        )
    )
    scan_id = scan_result.get("data").get("createWebScan").get("scan").get("id")
    return scan_id


def _create_test_credentials(test_credentials, runner):
    credential_ids = []
    for test_credential in test_credentials:
        test_credential_result = runner.execute(
            test_credentials_create_api.CreateTestCredentialAPIRequest(
                test_credential=test_credential
            )
        )
        credential_ids.append(
            test_credential_result.get("data")
            .get("createTestCredentials")
            .get("testCredentials")
            .get("id")
        )
    return credential_ids


@run.run.command()
@click.option("--url", help="List of Urls to scan.", required=True, multiple=True)
@click.pass_context
def link(ctx: click.core.Context, url: List[str]):
    """Run scan for links."""
    run_link_scan(ctx, url)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/use/device/device.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7ostorlab/agent/message/proto/v3/use/device/device.proto\x12*ostorlab.agent.message.proto.v3.use.device\"3\n\rLoginPassword\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\"1\n\x06SSHKey\x12\x13\n\x0bprivate_key\x18\x01 \x01(\t\x12\x12\n\npublic_key\x18\x02 \x01(\t\"\xb8\x01\n\x0b\x43redentials\x12S\n\x0elogin_password\x18\x01 \x01(\x0b\x32\x39.ostorlab.agent.message.proto.v3.use.device.LoginPasswordH\x00\x12\x45\n\x07ssh_key\x18\x02 \x01(\x0b\x32\x32.ostorlab.agent.message.proto.v3.use.device.SSHKeyH\x00\x42\r\n\x0b\x63redentials\":\n\x0b\x44\x65viceRelay\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\x11\"*\n\x08\x44uration\x12\r\n\x05nanos\x18\x01 \x01(\x05\x12\x0f\n\x07seconds\x18\x02 \x01(\x03\"H\n\x08\x41rgument\x12\x10\n\x08\x61rg_name\x18\x01 \x01(\t\x12\x11\n\targ_value\x18\x02 \x03(\t\x12\x17\n\x0f\x61rg_value_bytes\x18\x03 \x03(\x0c\"\xe9\x03\n\x07Message\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_type\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x03 \x01(\t\x12\x16\n\x0e\x64\x65vice_version\x18\x04 \x01(\t\x12S\n\x12\x64\x65vice_credentials\x18\x05 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.v3.use.device.Credentials\x12M\n\x0c\x64\x65vice_relay\x18\x06 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.v3.use.device.DeviceRelay\x12Y\n\x18\x64\x65vice_relay_credentials\x18\x07 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.v3.use.device.Credentials\x12\x46\n\x08\x64uration\x18\x08 \x01(\x0b\x32\x34.ostorlab.agent.message.proto.v3.use.device.Duration\x12\x42\n\x04\x61rgs\x18\t \x03(\x0b\x32\x34.ostorlab.agent.message.proto.v3.use.device.Argument')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.use.device.device_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LOGINPASSWORD._serialized_start=103
  _LOGINPASSWORD._serialized_end=154
  _SSHKEY._serialized_start=156
  _SSHKEY._serialized_end=205
  _CREDENTIALS._serialized_start=208
  _CREDENTIALS._serialized_end=392
  _DEVICERELAY._serialized_start=394
  _DEVICERELAY._serialized_end=452
  _DURATION._serialized_start=454
  _DURATION._serialized_end=496
  _ARGUMENT._serialized_start=498
  _ARGUMENT._serialized_end=570
  _MESSAGE._serialized_start=573
  _MESSAGE._serialized_end=1062
# @@protoc_insertion_point(module_scope)
